from autotrader.brokers.trading import Order

class Strategy:
    def __init__(self, params, data, pair, **kwargs):
        """Define all indicators used in the strategy.
        """
        self.name = "Template Strategy"
        self.data = data
        self.params = params
        self.pair = pair
        
        # Define any indicators used in the strategy
        
        # Construct indicators dict for plotting
        self.indicators = {'Indicator Name': {'type': 'indicatortype',
                                              'data': 'indicatordata'},
                           }
        
        
    def generate_signal(self, data):
        """Define strategy logic to determine entry signals.
        """
        
        # Example long market order
        order = Order(direction=1)
        
        return order
    


import os
import importlib
import numpy as np
import pandas as pd
from autotrader.comms import emailing
from datetime import datetime, timezone
from autotrader.autodata import GetData
from autotrader.brokers.trading import Order
from autotrader.utilities import read_yaml, get_config


class AutoTraderBot:
    """AutoTrader Trading Bot.
    
    Attributes
    ----------
    instrument : str
        The trading instrument assigned to the bot.
    data : pd.DataFrame
        The OHLC price data used by the bot.
    quote_data : pd.DataFrame
        The OHLC quote data used by the bot.
    MTF_data : dict
        The multiple timeframe data used by the bot.
    backtest_summary : dict
        A dictionary containing results from the bot in backtest. This 
        dictionary is available only after a backtest and has keys: 'data', 
        'account_history', 'trade_summary', 'indicators', 'instrument', 
        'interval', 'open_trades', 'cancelled_trades'.
    
    """
    
    def __init__(self, instrument: str, strategy_dict: dict, 
                 broker, deploy_dt: datetime, data_dict: dict, 
                 quote_data_path: str, auxdata: dict, 
                 autotrader_instance) -> None:
        """Instantiates an AutoTrader Bot.

        Parameters
        ----------
        instrument : str
            The trading instrument assigned to the bot instance.
        strategy_dict : dict
            The strategy configuration dictionary.
        broker : AutoTrader Broker instance
            The AutoTrader Broker module.
        deploy_dt : datetime
            The datetime stamp of the bot deployment time.
        data_dict : dict
            The strategy data.
        quote_data_path : str
            The quote data filepath for the trading instrument 
            (for backtesting only).
        auxdata : dict
            Auxiliary strategy data.
        autotrader_instance : AutoTrader
            The parent AutoTrader instance.

        Raises
        ------
        Exception
            When there is an error retrieving the instrument data.

        Returns
        -------
        None
            The trading bot will be instantiated and ready for trading.

        """
        # Inherit user options from autotrader
        for attribute, value in autotrader_instance.__dict__.items():
            setattr(self, attribute, value)
        self._scan_results = {}
        
        # Assign local attributes
        self.instrument = instrument
        self._broker = broker
        
        # Unpack strategy parameters and assign to strategy_params
        strategy_config = strategy_dict['config']
        interval = strategy_config["INTERVAL"]
        period = strategy_config["PERIOD"]
        risk_pc = strategy_config["RISK_PC"] if 'RISK_PC' \
            in strategy_config else None
        sizing = strategy_config["SIZING"] if 'SIZING' \
            in strategy_config else None
        params = strategy_config["PARAMETERS"]
        strategy_params = params
        strategy_params['granularity'] = strategy_params['granularity'] \
            if 'granularity' in strategy_params else interval
        strategy_params['risk_pc'] = strategy_params['risk_pc'] \
            if 'risk_pc' in strategy_params else risk_pc
        strategy_params['sizing'] = strategy_params['sizing'] \
            if 'sizing' in strategy_params else sizing
        strategy_params['period'] = strategy_params['period'] \
            if 'period' in strategy_params else period
        strategy_params['INCLUDE_POSITIONS'] = strategy_config['INCLUDE_POSITIONS'] \
            if 'INCLUDE_POSITIONS' in strategy_config else False
        strategy_config['INCLUDE_BROKER'] = strategy_config['INCLUDE_BROKER'] \
            if 'INCLUDE_BROKER' in strategy_config else False
        self._strategy_params = strategy_params
        
        # Import Strategy
        if strategy_dict['class'] is not None:
            strategy = strategy_dict['class']
        else:
            strat_module = strategy_config["MODULE"]
            strat_name = strategy_config["CLASS"]
            strat_package_path = os.path.join(self._home_dir, "strategies") 
            strat_module_path = os.path.join(strat_package_path, 
                                             strat_module) + '.py'
            strat_spec = importlib.util.spec_from_file_location(strat_module, 
                                                                strat_module_path)
            strategy_module = importlib.util.module_from_spec(strat_spec)
            strat_spec.loader.exec_module(strategy_module)
            strategy = getattr(strategy_module, strat_name)
        
        # Strategy shutdown routine
        self._strategy_shutdown_method = strategy_dict['shutdown_method']
        
        # Get broker configuration 
        if self._global_config_dict is not None:
            # Use global config dict provided
            global_config = self._global_config_dict
        else:
            global_config_fp = os.path.join(self._home_dir, 'config', 'GLOBAL.yaml')
            if os.path.isfile(global_config_fp):
                global_config = read_yaml(global_config_fp)
            else:
                global_config = None
        broker_config = get_config(self._environment, global_config, self._feed)
   
        # Data retrieval
        self._quote_data_file = quote_data_path     # Either str or None
        self._data_filepaths = data_dict            # Either str or dict, or None
        self._auxdata_files = auxdata               # Either str or dict, or None
        
        # Fetch data
        self._get_data = GetData(broker_config, self._allow_dancing_bears,
                                 self._base_currency)
        
        # Create instance of data stream object
        stream_attributes = {"data_filepaths": self._data_filepaths,
                             "quote_data_file": self._quote_data_file,
                             "auxdata_files": self._auxdata_files,
                             "strategy_params": self._strategy_params,
                             "get_data": self._get_data,
                             "data_start": self._data_start,
                             "data_end": self._data_end,
                             "instrument": self.instrument,
                             "feed": self._feed}
        self.Stream = self._data_stream_object(**stream_attributes)
        
        # Initial data call
        self._refresh_data(deploy_dt)
        
        # Instantiate Strategy
        # TODO - data stream is passed in here...
        if strategy_config['INCLUDE_BROKER']:
            my_strat = strategy(params, self._strat_data, instrument, 
                                self._broker, self._broker_utils, 
                                data_stream=self.Stream)
        else:
            my_strat = strategy(params, self._strat_data, instrument)
            
        # Assign strategy to local attributes
        self._last_bars = None
        self._strategy = my_strat
        self._strategy_name = strategy_config['NAME'] if 'NAME' in \
            strategy_config else '(unnamed strategy)'
        
        # Assign strategy attributes for tick-based strategy development
        if self._backtest_mode:
            self._strategy._backtesting = True
            self.backtest_summary = None
        if interval.split(',')[0] == 'tick':
            self._strategy._tick_data = True
        
        if int(self._verbosity) > 0:
                print(f"\nAutoTraderBot assigned to trade {instrument}",
                      f"with {self._broker_name} broker using {strategy_config['NAME']}.")
    
    
    def __repr__(self):
        return f'{self.instrument} AutoTraderBot'
    
    
    def __str__(self):
        return 'AutoTraderBot instance'
    
    
    def _update(self, i: int = None, timestamp: datetime = None) -> None:
        """Update strategy with the latest data and generate a trade signal.
        

        Parameters
        ----------
        i : int, optional
            The indexing parameter used when running in periodic update mode. 
            The default is None.
        timestamp : datetime, optional
            The timestamp parameter used when running in continuous update 
            mode. The default is None.

        Returns
        -------
        None
            Trade signals generated will be submitted to the broker.

        """
        
        if self._run_mode == 'continuous':
            # Running in continuous update mode
            strat_data, current_bars, quote_bars, sufficient_data = self._check_data(timestamp, self._data_indexing)
            strat_object = strat_data
            
        else:
            # Running in periodic update mode
            current_bars = {self.instrument: self.data.iloc[i]}
            quote_bars = {self.instrument: self.quote_data.iloc[i]}
            sufficient_data = True
            strat_object = i
        
        # Check for new data
        new_data = self._check_last_bar(current_bars) 
        
        if sufficient_data and new_data:
            if self._backtest_mode or self._virtual_livetrading:
                # Update virtual broker with latest price bars
                self._update_virtual_broker(current_bars)
            
            # Get strategy orders
            if self._strategy_params['INCLUDE_POSITIONS']:
                current_position = self._broker.get_positions(self.instrument)
                strategy_orders = self._strategy.generate_signal(strat_object, 
                                            current_position=current_position)
            else:
                strategy_orders = self._strategy.generate_signal(strat_object)
            
            # Check and qualify orders
            orders = self._check_orders(strategy_orders)
            self._qualify_orders(orders, current_bars, quote_bars)
            
            # Submit orders
            for order in orders:
                if self._scan_mode:
                    # Bot is scanning
                    scan_hit = {"size"  : order.size,
                                "entry" : current_bars[order.instrument].Close,
                                "stop"  : order.stop_loss,
                                "take"  : order.take_profit,
                                "signal": order.direction}
                    self._scan_results[self.instrument] = scan_hit
                    
                else:
                    # Bot is trading
                    try:
                        order_time = current_bars[order.instrument].name
                    except:
                        order_time = current_bars[order.data_name].name
                        
                    self._broker.place_order(order, order_time=order_time)
            
            if int(self._verbosity) > 1:
                current_time = current_bars[list(current_bars.keys())[0]].name.strftime('%b %d %Y %H:%M:%S')
                if len(orders) > 0:
                    for order in orders:
                        direction = 'long' if order.direction > 1 else 'short'
                        order_string = f"{current_time}: {order.instrument} "+\
                            f"{direction} {order.order_type} order of " + \
                            f"{order.size} units placed at {order.order_price}."
                        print(order_string)
                else:
                    if int(self._verbosity) > 2:
                        print(f"{current_time}: No signal detected ({self.instrument}).")
            
            # Check for orders placed and/or scan hits
            if int(self._notify) > 0 and not self._backtest_mode:
                for order in orders:
                    self._broker_utils.write_to_order_summary(order, 
                                                              self._order_summary_fp)
                
                if int(self._notify) > 1 and \
                    self._email_params['mailing_list'] is not None and \
                    self._email_params['host_email'] is not None:
                        if int(self._verbosity) > 0 and len(self._latest_orders) > 0:
                                print("Sending emails ...")
                                
                        for order in orders:
                            emailing.send_order(order,
                                                self._email_params['mailing_list'],
                                                self._email_params['host_email'])
                            
                        if int(self._verbosity) > 0 and len(orders) > 0:
                            print("  Done.\n")
            
            # Check scan results
            if self._scan_mode:
                # Construct scan details dict
                scan_details = {'index': self._scan_index,
                                'strategy': self._strategy.name,
                                'timeframe': self._strategy_params['granularity']
                                }
                
                # Report AutoScan results
                # Scan reporting with no emailing requested.
                if int(self._verbosity) > 0 or \
                    int(self._notify) == 0:
                    if len(self._scan_results) == 0:
                        print("{}: No signal detected.".format(self.instrument))
                    else:
                        # Scan detected hits
                        for instrument in self._scan_results:
                            signal = self._scan_results[instrument]['signal']
                            signal_type = 'Long' if signal == 1 else 'Short'
                            print(f"{instrument}: {signal_type} signal detected.")
                
                if int(self._notify) > 0:
                    # Emailing requested
                    if len(self._scan_results) > 0 and \
                        self._email_params['mailing_list'] is not None and \
                        self._email_params['host_email'] is not None:
                        # There was a scanner hit and email information is provided
                        emailing.send_scan_results(self._scan_results, 
                                                   scan_details, 
                                                   self._email_params['mailing_list'],
                                                   self._email_params['host_email'])
                    elif int(self._notify) > 1 and \
                        self._email_params['mailing_list'] is not None and \
                        self._email_params['host_email'] is not None:
                        # There was no scan hit, but notify set > 1, so send email
                        # regardless.
                        emailing.send_scan_results(self._scan_results, 
                                                   scan_details, 
                                                   self._email_params['mailing_list'],
                                                   self._email_params['host_email'])
        
    
    def _refresh_data(self, timestamp: datetime = None, **kwargs):
        """Refreshes the active Bot's data attributes for trading.

        Parameters
        ----------
        timestamp : datetime, optional
            The current timestamp. If None, datetime.now() will be called.
            The default is None.
        **kwargs : dict
            Any other named arguments.

        Raises
        ------
        Exception
            When there is an error retrieving the data.

        Returns
        -------
        None: 
            The up-to-date data will be assigned to the Bot instance.

        """
        
        timestamp = datetime.now(timezone.utc) if timestamp is None else timestamp
        
        # Fetch new data
        data, multi_data, quote_data, auxdata = self.Stream.refresh(timestamp=timestamp)
        
        # Check data returned is valid
        if len(data) == 0:
            raise Exception("Error retrieving data.")
        
        # Data assignment
        if multi_data is None:
            strat_data = data
        else:
            strat_data = multi_data
        
        # Auxiliary data assignment
        if auxdata is not None:
            strat_data = {'base': strat_data,
                          'aux': auxdata}
        
        # Assign data attributes to bot
        self._strat_data = strat_data
        self.data = data
        self.multi_data = multi_data
        self.auxdata = auxdata
        self.quote_data = quote_data
        
    
    def _check_orders(self, orders) -> list:
        """Checks that orders returned from strategy are in the correct
        format.
        
        Returns
        -------
        List of Orders
        
        Notes
        -----
        An order must have (at the very least) an order type specified. Usually,
        the direction will also be required, except in the case of close order 
        types. If an order with no order type is provided, it will be ignored.
        """
        
        def check_type(orders):
            checked_orders = []
            if isinstance(orders, dict):
                # Order(s) provided in dictionary
                if 'order_type' in orders:
                    # Single order dict provided
                    if 'instrument' not in orders:
                        orders['instrument'] = self.instrument
                    checked_orders.append(Order._from_dict(orders))
                    
                elif len(orders) > 0:
                    # Multiple orders provided
                    for key, item in orders.items():
                        if isinstance(item, dict) and 'order_type' in item:
                            # Convert order dict to Order object
                            if 'instrument' not in item:
                                item['instrument'] = self.instrument
                            checked_orders.append(Order._from_dict(item))
                        elif isinstance(item, Order):
                            # Native Order object, append as is
                            checked_orders.append(item)
                        else:
                            raise Exception(f"Invalid order submitted: {item}")
                
                elif len(orders) == 0:
                    # Empty order dict
                    pass
                
            elif isinstance(orders, Order):
                # Order object directly returned
                checked_orders.append(orders)
                
            elif isinstance(orders, list):
                # Order(s) provided in list
                for item in orders:
                    if isinstance(item, dict) and 'order_type' in item:
                        # Convert order dict to Order object
                        if 'instrument' not in item:
                            item['instrument'] = self.instrument
                        checked_orders.append(Order._from_dict(item))
                    elif isinstance(item, Order):
                        # Native Order object, append as is
                        checked_orders.append(item)
                    else:
                        raise Exception(f"Invalid order submitted: {item}")
            else:
                raise Exception(f"Invalid order submitted: {item}")
            
            return checked_orders
        
        def add_strategy_data(orders):
            # Append strategy parameters to each order
            for order in orders:
                order.instrument = self.instrument if not order.instrument else order.instrument
                order.strategy = self._strategy.name if 'name' in \
                    self._strategy.__dict__ else self._strategy_name
                order.granularity = self._strategy_params['granularity']
                order._sizing = self._strategy_params['sizing']
                order._risk_pc = self._strategy_params['risk_pc']
                
        def check_order_details(orders: list) -> None:
            for ix, order in enumerate(orders):
                order.instrument = order.instrument if order.instrument is not None else self.instrument
                if order.order_type in ['market', 'limit', 'stop-limit', 'reduce']:
                    if not order.direction:
                        del orders[ix]
        
        # Perform checks
        checked_orders = check_type(orders)
        add_strategy_data(checked_orders)
        check_order_details(checked_orders)
        
        return checked_orders
        
    
    def _qualify_orders(self, orders: list, current_bars: dict,
                        quote_bars: dict) -> None:
        """Passes price data to order to populate missing fields.
        """
        for order in orders:
            if self._req_liveprice:
                liveprice_func = getattr(self._get_data, f'{self._feed.lower()}_liveprice')
                last_price = liveprice_func(order)
            else:
                try:
                    last_price = self._get_data._pseduo_liveprice(last=current_bars[order.instrument].Close,
                                                                  quote_price=quote_bars[order.instrument].Close)
                except:
                    last_price = self._get_data._pseduo_liveprice(last=current_bars[order.data_name].Close,
                                                                  quote_price=quote_bars[order.data_name].Close)
            
            if order.order_type not in ['close', 'reduce']:
                if order.direction < 0:
                    order_price = last_price['bid']
                    HCF = last_price['negativeHCF']
                else:
                    order_price = last_price['ask']
                    HCF = last_price['positiveHCF']
            else:
                # Close or reduce order type, provide dummy inputs
                order_price = last_price['ask']
                HCF = last_price['positiveHCF']
            
            # Call order with price and time
            order(broker=self._broker, order_price=order_price, HCF=HCF)
    
    
    def _update_virtual_broker(self, current_bars: dict) -> None:
        """Updates virtual broker with latest price data.
        """
        for product, bar in current_bars.items():
            self._broker._update_positions(bar, product)
    
    
    def _create_backtest_summary(self, balance: pd.Series, NAV: pd.Series, 
                                margin: pd.Series, trade_times = None) -> dict:
        """Constructs backtest summary dictionary for further processing.
        """
        trade_summary = self._broker_utils.trade_summary(trades=self._broker.trades,
                                                         instrument=self.instrument)
        order_summary = self._broker_utils.trade_summary(orders=self._broker.orders,
                                                         instrument=self.instrument)
        
        if trade_times is None:
            trade_times = self.data.index
        
        # closed_trades = trade_summary[trade_summary.status == 'closed']
        open_trade_summary = trade_summary[trade_summary.status == 'open']
        cancelled_summary = order_summary[order_summary.status == 'cancelled']
        
        backtest_dict = {}
        backtest_dict['data'] = self.data
        backtest_dict['account_history'] = pd.DataFrame(data={'balance': balance, 
                                                              'NAV': NAV, 
                                                              'margin': margin,
                                                              'drawdown': np.array(NAV)/np.maximum.accumulate(NAV) - 1}, 
                                                        index=trade_times)
        backtest_dict['trade_summary'] = trade_summary
        backtest_dict['indicators'] = self._strategy.indicators if hasattr(self._strategy, 'indicators') else None
        backtest_dict['instrument'] = self.instrument
        backtest_dict['interval'] = self._strategy_params['granularity']
        backtest_dict['open_trades'] = open_trade_summary
        backtest_dict['cancelled_trades'] = cancelled_summary
        
        self.backtest_summary = backtest_dict
    
    
    def _get_iteration_range(self) -> int:
        """Checks mode of operation and returns data iteration range. For backtesting,
        the entire dataset is iterated over. For livetrading, only the latest candle
        is used. ONLY USED IN BACKTESTING NOW.
        """
        
        start_range = self._strategy_params['period']
        end_range = len(self.data)
        
        if len(self.data) < start_range:
            raise Exception("There are not enough bars in the data to " + \
                            "run the backtest with the current strategy " + \
                            "configuration settings. Either extend the " + \
                            "backtest period, or reduce the PERIOD key of " + \
                            "your strategy configuration.")
        
        return start_range, end_range
    
    
    @staticmethod
    def _check_ohlc_data(ohlc_data: pd.DataFrame, timestamp: datetime, 
                         indexing: str = 'open', tail_bars: int = None,
                         check_for_future_data: bool = True) -> pd.DataFrame:
        """Checks the index of inputted data to ensure it contains no future 
        data.

        Parameters
        ----------
        ohlc_data : pd.DataFrame
            DESCRIPTION.
        timestamp : datetime
            The current timestamp.
        indexing : str, optional
            How the OHLC data has been indexed (either by bar 'open' time, or
            bar 'close' time). The default is 'open'.
        tail_bars : int, optional
            If provided, the data will be truncated to provide the number
            of bars specified. The default is None.
        check_for_future_data : bool, optional
            A flag to check for future entries in the data. The default is True.
        
        Raises
        ------
        Exception
            When an unrecognised data indexing type is specified.

        Returns
        -------
        past_data : pd.DataFrame
            The checked data.

        """
        if check_for_future_data:
            if indexing.lower() == 'open':
                past_data = ohlc_data[ohlc_data.index < timestamp]
            elif indexing.lower() == 'close':
                past_data = ohlc_data[ohlc_data.index <= timestamp]
            else:
                raise Exception(f"Unrecognised indexing type '{indexing}'.")
        else:
            past_data = ohlc_data
        
        if tail_bars is not None:
            past_data = past_data.tail(tail_bars)
            
        return past_data
    
    
    def _check_auxdata(self, auxdata: dict, timestamp: datetime, 
                       indexing: str = 'open', tail_bars: int = None,
                       check_for_future_data: bool = True) -> dict:
        """Function to check the strategy auxiliary data.

        Parameters
        ----------
        auxdata : dict
            The strategy's auxiliary data.
        timestamp : datetime
            The current timestamp.
        indexing : str, optional
            How the OHLC data has been indexed (either by bar 'open' time, or
            bar 'close' time). The default is 'open'.
        tail_bars : int, optional
            If provided, the data will be truncated to provide the number
            of bars specified. The default is None.
        check_for_future_data : bool, optional
            A flag to check for future entries in the data. The default is True.

        Returns
        -------
        dict
            The checked auxiliary data.

        """
        processed_auxdata = {}
        for key, item in auxdata.items():
            if isinstance(item, pd.DataFrame) or isinstance(item, pd.Series):
                processed_auxdata[key] = self._check_ohlc_data(item, timestamp, 
                                    indexing, tail_bars, check_for_future_data)
            else:
                processed_auxdata[key] = item
        return processed_auxdata
                
    
    def _check_data(self, timestamp: datetime, indexing: str = 'open') -> dict:
        """Wrapper for multiple datasets contained in a dictionary.

        Parameters
        ----------
        timestamp : datetime
            DESCRIPTION.
        indexing : str, optional
            DESCRIPTION. The default is 'open'.

        Returns
        -------
        strat_data : dict
            The checked strategy data.
        current_bars : dict(pd.core.series.Series)
            The current bars for each product.
        quote_bars : dict(pd.core.series.Series)
            The current quote data bars for each product.
        sufficient_data : bool
            Boolean flag whether sufficient data is available.

        """
        # TODO - add check of data frequency against update frequency
        
        def get_current_bars(data: pd.DataFrame, quote_data: bool = False,
                             processed_strategy_data: dict = None) -> dict:
            """Returns the current bars of data. If the inputted data is for
            quote bars, then the quote_data boolean will be True.
            """
            if len(data) > 0:
                current_bars = self.Stream.get_trading_bars(data=data, 
                                quote_bars=quote_data, timestamp=timestamp,
                                processed_strategy_data=processed_strategy_data)
            else:
                current_bars = None
            return current_bars
        
        def process_strat_data(original_strat_data, check_for_future_data):
            sufficient_data = True
            
            if isinstance(original_strat_data, dict):
                if 'aux' in original_strat_data:
                    base_data = original_strat_data['base']
                    processed_auxdata = self._check_auxdata(original_strat_data['aux'],
                                    timestamp, indexing, bars, check_for_future_data)
                else:
                    # MTF data
                    base_data = original_strat_data
                
                # Process base OHLC data
                processed_basedata = {}
                for granularity, data in base_data.items():
                    processed_basedata[granularity] = self._check_ohlc_data(data, 
                                timestamp, indexing, bars, check_for_future_data)
                
                # Combine the results of the conditionals above
                strat_data = {}
                if 'aux' in original_strat_data:
                    strat_data['aux'] = processed_auxdata
                    strat_data['base'] = processed_basedata
                else:
                    strat_data = processed_basedata
                    
                # Extract current bar
                first_tf_data = processed_basedata[list(processed_basedata.keys())[0]]
                current_bars = get_current_bars(first_tf_data, 
                                                processed_strategy_data=strat_data)
                
                # Check that enough bars have accumulated
                if len(first_tf_data) < bars:
                    sufficient_data = False
                
            elif isinstance(original_strat_data, pd.DataFrame):
                strat_data = self._check_ohlc_data(original_strat_data, 
                             timestamp, indexing, bars, check_for_future_data)
                current_bars = get_current_bars(strat_data,
                                                processed_strategy_data=strat_data)
                
                # Check that enough bars have accumulated
                if len(strat_data) < bars:
                    sufficient_data = False
            
            else:
                raise Exception("Unrecognised data type. Cannot process.")
            
            return strat_data, current_bars, sufficient_data
        
        # Define minimum number of bars for strategy to run
        bars = self._strategy_params['period']
        
        if self._backtest_mode:
            check_for_future_data = True
            if self._dynamic_data:
                self._refresh_data(timestamp)
        else:
            # Livetrading
            self._refresh_data(timestamp)
            check_for_future_data = False
            
        strat_data, current_bars, sufficient_data = process_strat_data(self._strat_data, 
                                                                       check_for_future_data)

        # Process quote data
        quote_data = self._check_ohlc_data(self.quote_data, timestamp, 
                                           indexing, bars)
        quote_bars = get_current_bars(quote_data, True, strat_data)
        
        return strat_data, current_bars, quote_bars, sufficient_data
    
    
    def _check_last_bar(self, current_bars: dict) -> bool:
        """Checks for new data to prevent duplicate signals.
        """
        try:
            duplicated_bars = []
            for product, bar in current_bars.items():
                if (bar == self._last_bars[product]).all():
                    duplicated_bars.append(True)
                else:
                    duplicated_bars.append(False)
                    
            if len(duplicated_bars) == sum(duplicated_bars):
                new_data = False
            else: 
                new_data = True
            
        except:
            new_data = True
        
        # Reset last bars
        self._last_bars = current_bars
            
        if int(self._verbosity) > 1 and not new_data:
            print("Duplicate bar detected. Skipping.")
        
        return new_data
        
    
    def _check_strategy_for_plot_data(self, use_strat_plot_data: bool = False):
        """Checks the bot's strategy to see if it has the plot_data attribute.

        Returns
        -------
        plot_data : pd.DataFrame
            The data to plot.
            
        Notes
        -----
        This method is a placeholder for a future feature, allowing 
        customisation of what is plotted by setting plot_data and plot_type
        attributes from within a strategy.
        """
        strat_params = self._strategy.__dict__
        if 'plot_data' in strat_params and use_strat_plot_data:
            plot_data = strat_params['plot_data']
        else:
            plot_data = self.data
            
        return plot_data
    
    
    def _strategy_shutdown(self,):
        if self._strategy_shutdown_method is not None:
            try:
                shutdown_method = getattr(self._strategy, self._strategy_shutdown_method)
                shutdown_method()
            except AttributeError:
                print(f"\nShutdown method '{self._strategy_shutdown_method}' not found!")
            


import os
import pandas as pd
import autotrader.indicators as indicators

class TestIndicators:
    
    def test_supertrend(self):
        testdata = self._get_test_data()
        st = indicators.supertrend(testdata)
        assert st['trend'][-1] == 1, "Supertrend indicator failed"
    
    
    def test_halftrend(self):
        testdata = self._get_test_data()
        ht = indicators.halftrend(testdata)
        assert round(ht['halftrend'][-1], 5) == 1.12344, "Halftrend indicator failed"
        assert round(ht['atrHigh'][-1], 5) == 1.12593, "Halftrend indicator failed"
        assert round(ht['atrLow'][-1], 5) == 1.12095, "Halftrend indicator failed"
    
    
    def test_range_filter(self):
        testdata = self._get_test_data()
        rfi = indicators.range_filter(testdata)
        assert round(rfi['upper'][-1], 6) == 1.12736, "Range filter indicator failed"
        assert round(rfi['lower'][-1], 6) == 1.119815, "Range filter indicator failed"
        assert round(rfi['rf'][-1], 6) == 1.123588, "Range filter indicator failed"
    
    
    @staticmethod
    def test_rolling_signal_list():
        signals = [0,1,0,0,0,-1,0,0,1,0,0]
        rolled_signals = indicators.rolling_signal_list(signals)
        assert rolled_signals == [0, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1], "Rolling signal list indicator failed"
    
    
    @staticmethod
    def test_unroll_signal_list():
        signals = [0, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1]
        unrolled_signals = indicators.unroll_signal_list(signals)
        assert list(unrolled_signals) == [0,1,0,0,0,-1,0,0,1,0,0], "Unroll signal list indicator failed"
    
    
    @staticmethod
    def _get_test_data():
        home_dir = os.path.abspath(os.path.dirname(__file__))
        data_dir = os.path.join(home_dir, 'data', 'EUR_USD_H4.csv')
        data = pd.read_csv(data_dir, index_col=0, 
                           parse_dates=True)
        testdata = data.iloc[-1000:]
        return testdata


from autotrader.brokers.trading import Order
from autotrader.brokers.broker_utils import BrokerUtils

"""
Notes:
    - Public methods are called from outside the broker module, and so must
      retain functionality of input arguments. If necessary, they can simply
      be wrapper methods.
    - Private methods are broker-specific.
"""

class Broker:
    def __init__(self, config: dict, utils: BrokerUtils = None) -> None:
        """AutoTrader Broker Class constructor.
        """
        
        self.utils = utils if utils is not None else BrokerUtils()
        
        # Unpack config and connect to broker-side API
        
    
    def __repr__(self):
        return 'AutoTrader Broker interface'
    
    
    def __str__(self):
        return 'AutoTrader Broker interface'
    
    
    def get_NAV(self) -> float:
        """Returns the net asset/liquidation value of the account.
        """
        pass
    
    
    def get_balance(self) -> float:
        """Returns account balance.
        """
        pass
        
    
    def place_order(self, order: Order, **kwargs) -> None:
        """Disassemble order_details dictionary to place order.
        """
        # Call order to set order time
        order()
        
        # Submit order to broker
        
    
    def get_orders(self, instrument: str = None, **kwargs) -> dict:
        """Returns all pending orders (have not been filled) in the account.
        """
        pass
    
    
    def cancel_order(self, order_id: int, **kwargs) -> None:
        """Cancels order by order ID.
        """
        pass
    
    
    def get_trades(self, instrument: str = None, **kwargs) -> dict:
        """Returns the open trades held by the account. 
        """
        pass
    
    
    def get_trade_details(self, trade_ID: str) -> dict:
        """Returns the details of the trade specified by trade_ID.
        """
        pass
    
    
    def get_positions(self, instrument: str = None, **kwargs) -> dict:
        """Gets the current positions open on the account.
        
        Parameters
        ----------
        instrument : str, optional
            The trading instrument name (symbol). The default is None.
            
        Returns
        -------
        open_positions : dict
            A dictionary containing details of the open positions.
        """
        pass
    
    
    # Define here any private methods to support the public methods above
    
    


import os
import sys
import time
import timeit
import pyfiglet
import importlib
import numpy as np
import pandas as pd
from ast import literal_eval
from scipy.optimize import brute
from autotrader.autoplot import AutoPlot
from autotrader.autobot import AutoTraderBot
from datetime import datetime, timedelta, timezone
from autotrader.utilities import read_yaml, get_config, get_watchlist, DataStream


class AutoTrader:
    """A Python-Based Development Platform For Automated Trading Systems.
    
    Methods
    -------
    configure(...)
        Configures run settings for AutoTrader.
    add_strategy(...)
        Adds a strategy to the active AutoTrader instance. 
    backtest(...)
        Configures backtest settings.
    optimise(...)
        Configures optimisation settings.
    scan(...)
        Configures scan settings.
    run()
        Runs AutoTrader with configured settings.
    add_data(...)
        Specify local data files to use for backtests.
    plot_settings(...)
        Configures the plot settings for AutoPlot.
    get_bots_deployed(instrument=None)
        Returns the AutoTrader trading bots deployed in the active instance.
    plot_backtest(bot=None)
        Plots backtest results of a trading Bot.
    plot_multibot_backtest(backtest_results=None)
        Plots backtest results for multiple trading bots.
    analyse_backtest(bot=None)
        Analyse backtest results of a single trading bot.
    multibot_backtest_analysis(bots=None)
        Analyses backtest results of multiple trading bots.
    print_backtest_results(backtest_results)
        Prints backtest results.
    print_multibot_backtest_results(backtest_results=None)
        Prints a multi-bot backtest results.
    
    References
    ----------
    Author: Kieran Mackle
    
    Version: 0.6.0
    
    Homepage: https://kieran-mackle.github.io/AutoTrader/
    
    GitHub: https://github.com/kieran-mackle/AutoTrader
    """
    
    def __init__(self) -> None:
        """AutoTrader initialisation. Called when creating new AutoTrader 
        instance.
        """
        
        self._home_dir = None
        self._verbosity = 1
        
        self._global_config_dict = None
        self._instance_str = None
        self._run_mode = 'periodic'
        self._timestep = pd.Timedelta('10s').to_pytimedelta()
        self._feed = 'yahoo'
        self._req_liveprice = False
        
        self._notify = 0
        self._email_params = None
        self._order_summary_fp = None
        self._show_plot = False
        
        # Livetrade Parameters
        self._check_data_alignment = True
        self._allow_dancing_bears = False
        
        self._broker = None
        self._broker_utils = None
        self._broker_verbosity = 0
        self._environment = 'demo'
        self._account_id = None
        
        self._strategy_configs = {}
        self._strategy_classes = {}
        self._shutdown_methods = {}
        self._uninitiated_strat_files = []
        self._uninitiated_strat_dicts = []
        self._bots_deployed = []
        
        # Backtesting Parameters
        self._backtest_mode = False
        self._data_start = None
        self._data_end = None
        
        # Local Data Parameters
        self._data_indexing = 'open'
        self._data_stream_object = DataStream
        self._data_file = None
        self._MTF_data_files = None
        self._local_data = None
        self._local_quote_data = None
        self._auxdata = None
        self._dynamic_data = False
        
        # Virtual Broker Parameters
        self._virtual_livetrading = False
        self._virtual_initial_balance = None
        self._virtual_spread = None
        self._virtual_commission = None
        self._virtual_leverage = None
        self._virtual_broker_hedging = False
        self._virtual_margin_call = 0
        self._base_currency = None
        
        # Optimisation Parameters
        self._optimise_mode = False
        self._opt_params = None
        self._bounds = None
        self._Ns = None
        
        # Scan Parameters
        self._scan_mode = False
        self._scan_index = None
        self._scan_watchlist = None
        
        # Plotting
        self._max_indis_over = 3
        self._max_indis_below = 2
        self._fig_tools = "pan,wheel_zoom,box_zoom,undo,redo,reset,save,crosshair"
        self._ohlc_height = 400
        self._ohlc_width = 800
        self._top_fig_height = 150
        self._bottom_fig_height = 150
        self._jupyter_notebook = False
        self._show_cancelled = True
        self._chart_timeframe = 'default'
        self._chart_theme = 'caliber'
        self._use_strat_plot_data = False
        
    
    def __repr__(self):
        return f'AutoTrader instance in {self._feed} {self._environment} environment'
    
    
    def __str__(self):
        return 'AutoTrader instance'
    
    
    def configure(self, verbosity: int = 1, broker: str = 'virtual', 
                  feed: str = 'yahoo', req_liveprice: bool = False, 
                  notify: int = 0, home_dir: str = None, 
                  allow_dancing_bears: bool = False, account_id: str = None, 
                  environment: str = 'demo', show_plot: bool = False,
                  jupyter_notebook: bool = False, mode: str = 'periodic',
                  update_interval: str = '10s', data_index_time: str = 'open',
                  global_config: dict = None, instance_str: str = None,
                  broker_verbosity: int = 0) -> None:
        """Configures run settings for AutoTrader.

        Parameters
        ----------
        verbosity : int, optional
            The verbosity of AutoTrader (0, 1, 2). The default is 1.
        broker : str, optional
            The broker to connect to. The default is 'virtual'.
        feed : str, optional
            The data feed to be used ('yahoo', 'oanda', 'ib'). The default is 'yahoo'.
        req_liveprice : bool, optional
            Request live market price from broker before placing trade, rather 
            than using the data already provided. The default is False. 
        notify : int, optional
            The level of email notifications (0, 1, 2). The default is 0.
        home_dir : str, optional
            The project home directory. The default is the current working directory.
        allow_dancing_bears : bool, optional
            Allow incomplete candles to be passed to the strategy. The default is False.
        account_id : str, optional
            The brokerage account ID to be used. The default is None.
        environment : str, optional
            The trading environment of this instance ('demo', 'real'). The 
            default is 'demo'.
        show_plot : bool, optional
            Automatically generate backtest chart. The default is False.
        jupyter_notebook : bool, optional
            Set to True when running in Jupyter notebook environment. The 
            default is False.
        mode : str, optional
            The run mode (either 'periodic' or 'continuous'). The default is
            'periodic'.
        update_interval : str, optional
            The update interval to use when running in 'continuous' mode. This
            should align with the highest resolution bar granularity in your
            strategy to allow adequate updates. The string inputted will be
            converted to a timedelta object. The default is '10s'.
        data_index_time : str, optional
            The time by which the data is indexed. Either 'open', if the data
            is indexed by the bar open time, or 'close', if the data is indexed
            by the bar close time. The default is 'open'.
        global_config : dict, optional
            Optionally provide your global configuration directly as a 
            dictionary, rather than it being read in from a yaml file. The
            default is None. 
        instance_str : str, optional
            The name of the active AutoTrader instance, used to control bots
            deployed when livetrading in continuous mode. When not specified,
            the instance string will be of the form 'autotrader_instance_n'.
            The default is None.
        broker_verbosity : int, optional
            The verbosity of the broker. The default is 0.
        
        Returns
        -------
        None
            Calling this method configures the internal settings of 
            the active AutoTrader instance.
        """
        self._verbosity = verbosity
        self._feed = feed
        self._req_liveprice = req_liveprice
        self._broker_name = broker
        self._notify = notify
        self._home_dir = home_dir if home_dir is not None else os.getcwd()
        self._allow_dancing_bears = allow_dancing_bears
        self._account_id = account_id
        self._environment = environment
        self._show_plot = show_plot
        self._jupyter_notebook = jupyter_notebook
        self._run_mode = mode
        self._timestep = pd.Timedelta(update_interval).to_pytimedelta()
        self._data_indexing = data_index_time
        self._global_config_dict = global_config
        self._instance_str = instance_str
        self._broker_verbosity = broker_verbosity
        
        
    def virtual_livetrade_config(self, initial_balance: float = 1000, 
                                  spread: float = 0, commission: float = 0, 
                                  leverage: int = 1, base_currency: str = 'AUD', 
                                  hedging: bool = False, 
                                  margin_call_fraction: float = 0) -> None:
        """Configures the virtual broker's initial state to allow livetrading
        on the virtual broker.
        
        Parameters
        ----------
        initial_balance : float, optional
            The initial balance of the account. The default is 1000.
        spread : float, optional
            The bid/ask spread to use. The default is 0.
        commission : float, optional
            Trading commission as percentage per trade. The default is 0.
        leverage : int, optional
            Account leverage. The default is 1.
        base_currency : str, optional
            The base currency of the account. The default is 'AUD'.
        hedging : bool, optional
            Allow hedging in the virtual broker (opening simultaneous 
            trades in oposing directions). The default is False.
        margin_call_fraction : float, optional
            The fraction of margin usage at which a margin call will occur.
            The default is 0.
            
        """
        # Assign attributes
        self._virtual_livetrading = True
        self._virtual_initial_balance = initial_balance
        self._virtual_spread = spread
        self._virtual_commission = commission
        self._virtual_leverage = leverage
        self._virtual_broker_hedging = hedging
        self._virtual_margin_call = margin_call_fraction
        self._base_currency = base_currency
        
        # Enforce virtual broker
        self._broker_name = 'virtual'
        

    def add_strategy(self, config_filename: str = None, 
                     config_dict: dict = None, strategy = None,
                     shutdown_method: str = None) -> None:
        """Adds a strategy to AutoTrader.

        Parameters
        ----------
        config_filename : str, optional
            The prefix of the yaml strategy configuration file, located in 
            home_dir/config. The default is None.
        config_dict : dict, optional
            Alternative to config_filename, a strategy configuration 
            dictionary can be passed directly. The default is None.
        strategy : AutoTrader Strategy, optional
            The strategy class object. The default is None.
        shutdown_method : str, optional
            The name of the shutdown method in the strategy (if any). This
            method will be called when AutoTrader is livetrading in continuous
            mode, and the instance has recieved the shutdown signal. The 
            default is None.

        Returns
        -------
        None
            The strategy will be added to the active AutoTrader instance.
        """
        
        if self._home_dir is None:
            # Home directory has not yet been set, postpone strategy addition
            if config_filename is None:
                self._uninitiated_strat_dicts.append(config_dict)
            else:
                self._uninitiated_strat_files.append(config_filename)
            
            if shutdown_method is not None:
                raise Exception("Providing the shutdown method requires "+\
                                "the home directory to have been configured. "+\
                                "please either specify it, or simply call "+\
                                "the configure method before adding a strategy.")
            
        else:
            # Home directory has been set
            if config_dict is None:
                config_file_path = os.path.join(self._home_dir, 'config', config_filename)
                new_strategy = read_yaml(config_file_path + '.yaml')
            else:
                new_strategy = config_dict
            
            name = new_strategy['NAME']
            
            if name in self._strategy_configs:
                print("Warning: duplicate strategy name deteced. Please check " + \
                      "the NAME field of your strategy configuration file and " + \
                      "make sure it is not the same as other strategies being " + \
                      "run from this instance.")
                print("Conflicting name:", name)
            
            self._strategy_configs[name] = new_strategy
            
            self._shutdown_methods[name] = shutdown_method
            
        if strategy is not None:
            self._strategy_classes[strategy.__name__] = strategy
            
    
    def backtest(self, start: str = None, end: str = None, 
                 initial_balance: float = 1000, spread: float = 0, 
                 commission: float = 0, leverage: int = 1,
                 base_currency: str = 'AUD', start_dt: datetime = None, 
                 end_dt: datetime = None, hedging: bool = False,
                 margin_call_fraction: float = 0) -> None:
        """Configures settings for backtesting.

        Parameters
        ----------
        start : str, optional
            Start date for backtesting, in format dd/mm/yyyy. The default is None.
        end : str, optional
            End date for backtesting, in format dd/mm/yyyy. The default is None.
        start_dt : datetime, optional
            Datetime object corresponding to start time. The default is None.
        end_dt : datetime, optional
            Datetime object corresponding to end time. The default is None.
        initial_balance : float, optional
            The initial balance of the account. The default is 1000.
        spread : float, optional
            The bid/ask spread to use in backtest. The default is 0.
        commission : float, optional
            Trading commission as percentage per trade. The default is 0.
        leverage : int, optional
            Account leverage. The default is 1.
        base_currency : str, optional
            The base currency of the account. The default is 'AUD'.
        hedging : bool, optional
            Allow hedging in the virtual broker (opening simultaneous 
            trades in oposing directions). The default is False.
        margin_call_fraction : float, optional
            The fraction of margin usage at which a margin call will occur.
            The default is 0.
            
        Notes
        ------
            Start and end times must be specified as the same type. For
            example, both start and end arguments must be provided together, 
            or alternatively, start_dt and end_dt must both be provided.
        """
        
        # Convert start and end strings to datetime objects
        if start_dt is None and end_dt is None:
            start_dt = datetime.strptime(start + '+0000', '%d/%m/%Y%z')
            end_dt = datetime.strptime(end + '+0000', '%d/%m/%Y%z')
        
        # Assign attributes
        self._backtest_mode = True
        self._data_start = start_dt
        self._data_end = end_dt
        self._virtual_initial_balance = initial_balance
        self._virtual_spread = spread
        self._virtual_commission = commission
        self._virtual_leverage = leverage
        self._virtual_broker_hedging = hedging
        self._virtual_margin_call = margin_call_fraction
        self._base_currency = base_currency
        
        # Enforce virtual broker
        self._broker_name = 'virtual'
        
    
    def optimise(self, opt_params: list, bounds: list, Ns: int = 4,
                 force_download: bool = False) -> None:
        """Optimisation configuration.
        
        Parameters
        ----------
        opt_params : list
            The parameters to be optimised, as they  are named in the 
            strategy configuration file.
        bounds : list(tuples)
            The bounds on each of the parameters to be optimised, specified
            as a tuple of the form (lower, upper) for each parameter. The
            default is 4.
        force_download : bool, optional
            Force AutoTrader to download data each iteration. This is not
            recommended. Instead, you should provide local download to optimise
            on, using the add_data method. The default is False.
        Ns : int, optional
            The number of points along each dimension of the optimisation grid.
        
        Raises
        ------
        Exception:
            When force_download is False, and local data has not been added 
            through the add_data method. Note that add_data should be called
            prior to calling the optimise method.
        
        Returns
        -------
        None:
            The optimisation settings will be saved to the active AutoTrader
            instance.
        
        See Also
        --------
        AutoTrader.add_data()
        """
        
        if type(bounds) == str:
            full_tuple = literal_eval(bounds)
            bounds = [(x[0], x[-1]) for x in full_tuple]

        if type(opt_params) == str:
            opt_params = opt_params.split(',')
        
        self._optimise_mode = True
        self._opt_params = opt_params
        self._bounds = bounds
        self._Ns = Ns
        
        if self._local_data is None:
            raise Exception("Local data files have not been provided. " +\
                            "Please do so using AutoTrader.add_data(), " +\
                            "or set force_download to True to proceed.")
        
        
    def add_data(self, data_dict: dict = None, quote_data: dict = None, 
                 data_directory: str = 'price_data', abs_dir_path: str = None, 
                 auxdata: dict = None, stream_object = None,
                 dynamic_data: bool = False) -> None:
        """Specify local data to run AutoTrader on.

        Parameters
        ----------
        data_dict : dict, optional
            A dictionary containing the filenames of the datasets
            to be used. The default is None.
        quote_data : dict, optional
            A dictionary containing the quote data filenames 
            of the datasets provided in data_dict. The default is None.
        data_directory : str, optional
            The name of the sub-directory containing price
            data files. This directory should be located in the project
            home directory (at.home_dir). The default is 'price_data'.
        abs_dir_path : str, optional
            The absolute path to the data_directory. This parameter
            may be used when the datafiles are stored outside of the project
            directory. The default is None.
        auxdata : dict, optional
            A dictionary containing the data paths to supplement the 
            data passed to the strategy module. For strategies involving 
            multiple products, the keys of this dictionary must correspond
            to the products, with the auxdata in nested dictionaries or 
            otherwise. The default is None.
        stream_object : DataStream, optional
            A custom data stream object, allowing custom data pipelines. The 
            default is DataStream (from autotrader.utilities).
        dynamic_data : bool, optional
            A boolean flag to signal that the stream object provided should 
            be refreshed each timestep of a backtest.
        
        Raises
        ------
        Exception
            When multiple quote-data files are provided per instrument traded.

        Returns
        -------
        None
            Data will be assigned to the active AutoTrader instance for
            later use.
        
        Notes
        ------
            To ensure proper directory configuration, this method should only 
            be called after calling autotrader.configure().
            
            The data provided to the strategy will either contain a single 
            timeframe OHLC dataframe, a dictionary of MTF dataframes, or 
            a dict with 'base' and 'aux' keys, for aux and base strategy 
            data (which could be single of MTF).
            
        Examples
        --------
            An example data_dict is shown below.
            
            >>> data_dict = {'product1': 'filename1.csv',
                             'product2': 'filename2.csv'}
            
            For MTF data, data_dict should take the form shown below. In 
            the case of MTF data, quote data should only be provided for
            the base timeframe (ie. the data which will be iterated on
            when backtesting). Therefore, the quote_data dict will look
            the same for single timeframe and MTF backtests.
            
            >>> data_dict = {'product1': {'H1': 'product1_H1.csv',
                                          'D': 'product1_D.csv'},
                             'product2': {'H1': 'product2_H1.csv',
                                          'D': 'product2_D.csv'}
                             }
        
            An example for the quate_data dictionary is shown below.
        
            >>> quote_data = {'product1': 'product1_quote.csv',
                              'product2': 'product2_quote.csv'}
        
            The auxdata dictionary can take the form shown below. This data 
            will be passed on to your strategy.
        
            >>> auxdata = {'product1': 'aux_price_data.csv', 
                           'product2': {'extra_data1': 'dataset1.csv',
                                        'extra_data2': 'dataset2.csv'}
                           }
            
        """
        dir_path = abs_dir_path if abs_dir_path is not None \
            else os.path.join(self._home_dir, data_directory)
        
        # Trading data
        if data_dict is not None:
            # Assign local data attribute
            local_data = {}
            
            # Populate local_data
            for product in data_dict:
                if type(data_dict[product]) == dict:
                    # MTF data
                    MTF_data = {}
                    for timeframe in data_dict[product]:
                        MTF_data[timeframe] = os.path.join(dir_path, data_dict[product][timeframe])
                    
                    local_data[product] = MTF_data
                else:
                    # Single timeframe data
                    local_data[product] = os.path.join(dir_path, data_dict[product])
                    
            self._local_data = local_data
        
        # Quote data
        if quote_data is not None:
            # Assign local data attribute
            local_quote_data = {}
            
            # Populate local_quote_data
            for product in quote_data:
                if type(quote_data[product]) == dict:
                    raise Exception("Only a single quote-data file should be " +\
                                    "provided per instrument traded.")
                    
                else:
                    local_quote_data[product] = os.path.join(dir_path, quote_data[product])
            
            self._local_quote_data = local_quote_data
        
        if auxdata is not None:
            modified_auxdata = {}
            for product, item in auxdata.items():
                if isinstance(item, dict):
                    # Multiple datasets for product
                    modified_auxdata[product] = {}
                    for key, dataset in item:
                        modified_auxdata[product][key] = os.path.join(dir_path, dataset)
                else:
                    # Item is not dict
                    modified_auxdata[product] = os.path.join(dir_path, item)
                    
            self._auxdata = modified_auxdata
        
        # Assign data stream object
        if stream_object is not None:
            self._data_stream_object = stream_object
        
        self._dynamic_data = dynamic_data
    
    
    def scan(self, strategy_filename: str = None, 
             strategy_dict: dict = None, scan_index: str = None) -> None:
        """Configure AutoTrader scan settings. 
        
        Parameters
        ----------
        strategy_filename : str, optional
             The prefix of yaml strategy configuration file, located in 
             home_dir/config. The default is None.
        strategy_dict : dict, optional
            A strategy configuration dictionary. The default is None.
        scan_index : str, optional
            Forex scan index. The default is None.

        Returns
        -------
        None
            The scan settings of the active AutoTrader instance will be
            configured.
        """
        
        # If a strategy is provided here, add it
        if strategy_filename is not None:
            self.add_strategy(strategy_filename)
        elif strategy_dict is not None:
            self.add_strategy(strategy_dict=strategy_dict)
        
        # If scan index provided, use that. Else, use strategy watchlist
        if scan_index is not None:
            self._scan_watchlist = get_watchlist(scan_index, self._feed)

        else:
            scan_index = 'Strategy watchlist'
            
        self._scan_mode = True
        self._scan_index = scan_index
        self._check_data_alignment = False
    
    
    def run(self) -> None:
        """Performs essential checks and runs AutoTrader.
        """
        # Define home_dir if undefined
        if self._home_dir is None:
            self._home_dir = os.getcwd()
        
        # Load uninitiated strategies
        for strat_dict in self._uninitiated_strat_dicts:
            self.add_strategy(strategy_dict=strat_dict)
        for strat_config_file in self._uninitiated_strat_files:
            self.add_strategy(strategy_filename=strat_config_file)
        
        if self._scan_watchlist is not None:
            # Scan watchlist has not overwritten strategy watchlist
            self._update_strategy_watchlist()
        
        if len(self._strategy_configs) == 0:
            print("Error: no strategy has been provided. Do so by using the" +\
                  " 'add_strategy' method of AutoTrader.")
            sys.exit(0)
            
        if sum([self._backtest_mode, self._scan_mode]) > 1:
            print("Error: backtest mode and scan mode are both set to True," +\
                  " but only one of these can run at a time.")
            print("Please check your inputs and try again.")
            sys.exit(0)
        
        if self._backtest_mode:
            if self._notify > 0:
                print("Warning: notify set to {} ".format(self._notify) + \
                      "during backtest. Setting to zero to prevent emails.")
                self._notify = 0
        
        if self._optimise_mode:
            if self._backtest_mode:
                self._run_optimise()
            else:
                print("Please set backtest parameters to run optimisation.")
        else:
            if not self._backtest_mode and self._broker_name == 'virtual':
                if not self._virtual_livetrading:
                    raise Exception("Live-trade mode requires setting the "+\
                                    "broker. Please do so using the "+\
                                    "AutoTrader configure method. If you "+\
                                    "would like to use the virtual broker "+\
                                    "for sandbox livetrading, please "+\
                                    "configure the virtual broker account "+\
                                    "with the virtual_livetrade_config method.")
                
            self._main()
    
    
    def plot_settings(self, max_indis_over: int = 3, max_indis_below: int = 2,
                      fig_tools: str = "pan,wheel_zoom,box_zoom,undo,redo,reset,save,crosshair",
                      ohlc_height: int = 400, ohlc_width: int = 800, 
                      top_fig_height: int = 150, bottom_fig_height: int = 150, 
                      jupyter_notebook: bool = False, show_cancelled: bool = True,
                      chart_timeframe: str = 'default', chart_theme: str = 'caliber',
                      use_strat_plot_data: bool = False) -> None:
        """Configure the plot settings.

        Parameters
        ----------
        max_indis_over : int, optional
            Maximum number of indicators overlaid on the main chart. The 
            default is 3.
        max_indis_below : int, optional
            Maximum number of indicators below the main chart. The default is 2.
        fig_tools : str, optional
            The figure tools. The default is "pan,wheel_zoom,box_zoom,undo,
            redo,reset,save,crosshair".
        ohlc_height : int, optional
            The height (px) of the main chart. The default is 400.
        ohlc_width : int, optional
            The width (px) of the main chart. The default is 800.
        top_fig_height : int, optional
            The height (px) of the figure above the main chart. The default is 150.
        bottom_fig_height : int, optional
            The height (px) of the figure(s) below the main chart. The default is 150.
        jupyter_notebook : bool, optional
            Boolean flag when running in Jupyter Notebooks, to allow inline
            plotting. The default is False.
        show_cancelled : bool, optional
            Show/hide cancelled trades. The default is True.
        chart_timeframe : str, optional
            The bar timeframe to use when gerating the chart. The timeframe
            provided must be a part of the strategy dataset. The default is 'default'.
        chart_theme : bool, optional
            The theme of the Bokeh chart generated. The default is "caliber".
        use_strat_plot_data : bool, optional
            Boolean flag to use data from the strategy instead of candlestick
            data for the chart. If True, ensure your strategy has a timeseries
            data attribute named 'plot_data'. The default is False.
        
        Returns
        -------
        None
            The plot settings will be saved to the active AutoTrader instance.
        """
        
        # Assign attributes
        self._max_indis_over    = max_indis_over
        self._max_indis_below   = max_indis_below
        self._fig_tools         = fig_tools
        self._ohlc_height       = ohlc_height
        self._ohlc_width        = ohlc_width
        self._top_fig_heigh     = top_fig_height
        self._bottom_fig_height = bottom_fig_height
        self._jupyter_notebook  = jupyter_notebook
        self._show_cancelled    = show_cancelled
        self._chart_timefram    = chart_timeframe
        self._chart_theme       = chart_theme
        self._use_strat_plot_data = use_strat_plot_data
    
    
    def get_bots_deployed(self, instrument: str = None) -> dict:
        """Returns a dictionary of AutoTrader trading bots, organised by 
        instrument traded.

        Parameters
        ----------
        instrument : str, optional
            The instrument of the bot to retrieve. The default is None.
            
        Returns
        -------
        dict
            A dictionary of deployed AutoTrader bot instances.
        
        Notes
        -----
        If there is only one trading bot deployed, this will be returned 
        directly, rather than in a dict.
        """
        bots = {}
        for bot in self._bots_deployed:
            bots[bot.instrument] = bot
        
        if instrument is not None:
            # Retrieve bot requested
            try:
                bots = bots[instrument]
            except:
                raise Exception(f"There were no bots found to be trading '{instrument}'.")
        
        if len(bots) == 1:
            # Single bot backtest, return directly
            bots = bots[list(bots.keys())[0]]
        
        return bots
        
    
    def plot_backtest(self, bot=None) -> None:
        """Plots backtest results of an AutoTrader Bot.
        
        Parameters
        ----------
        bot : AutoTrader bot instance, optional
            AutoTrader bot class containing backtest results. The default 
            is None.

        Returns
        -------
        None
            A chart will be generated and shown.
        """
        
        if bot is None:
            if len(self._bots_deployed) == 1:
                bot = self._bots_deployed[0]
            else:
                # Multi-bot backtest
                self.plot_multibot_backtest()
                return
        
        data = bot._check_strategy_for_plot_data(self._use_strat_plot_data)
        ap = self._instantiate_autoplot(data)
        profit_df = pd.merge(bot.data, 
                             bot.backtest_summary['trade_summary']['profit'], 
                             left_index=True, right_index=True).profit.cumsum()
        
        ap.plot(backtest_dict=bot.backtest_summary, cumulative_PL=profit_df)
    
    
    def plot_multibot_backtest(self) -> None:
        """Plots the backtest results for multiple trading bots.
        
        Returns
        -------
        None
            A chart will be generated and shown.
        """
        cpl_dict = {}
        for bot in self._bots_deployed:
            profit_df = pd.merge(bot.data, 
                     bot.backtest_summary['trade_summary']['profit'], 
                     left_index=True, right_index=True).profit.cumsum()
            cpl_dict[bot.instrument] = profit_df
        
        ap = self._instantiate_autoplot(bot.data)
        ap._plot_multibot_backtest(self.multibot_backtest_results, 
                                   bot.backtest_summary['account_history']['NAV'], 
                                   cpl_dict, 
                                   bot.backtest_summary['account_history']['margin'])
        
    
    def analyse_backtest(self, bot = None) -> dict:
        """Analyses bot backtest results to extract key statistics.
        
        Parameters
        ----------
        bot : AutoTraderBot, optional
            An AutoTraderBot class instance. The default is None.
            
        Returns
        -------
        backtest_results : dict
            A dictionary of backtest results.
        
        Notes
        -----
        If no bot is supplied, the backtest will be analysed for all bots 
        assiged during the backtest.
        
        See Also
        --------
        get_bots_deployed
        """
        
        if bot is None:
            if len(self._bots_deployed) == 1:
                bot = self._bots_deployed[0]
            else:
                print("Reverting to multi-bot backtest.")
                return self.multibot_backtest_analysis()
                    
        backtest_summary = bot.backtest_summary
        
        trade_summary   = backtest_summary['trade_summary']
        instrument      = backtest_summary['instrument']
        account_history = backtest_summary['account_history']
        
        cpl = trade_summary.profit.cumsum()
        
        backtest_results = {}
        
        # All trades
        no_trades = len(trade_summary[trade_summary['status'] == 'closed'])
        backtest_results['no_trades'] = no_trades
        backtest_results['start'] = account_history.index[0]
        backtest_results['end'] = account_history.index[-1]
        
        if no_trades > 0:
            backtest_results['all_trades'] = {}
            wins        = trade_summary[trade_summary.profit > 0]
            avg_win     = np.mean(wins.profit)
            max_win     = np.max(wins.profit)
            loss        = trade_summary[trade_summary.profit < 0]
            avg_loss    = abs(np.mean(loss.profit))
            max_loss    = abs(np.min(loss.profit))
            win_rate    = 100*len(wins)/no_trades
            longest_win_streak, longest_lose_streak  = self._broker_utils.get_streaks(trade_summary)
            avg_trade_duration = np.nanmean(trade_summary.trade_duration.values)
            min_trade_duration = np.nanmin(trade_summary.trade_duration.values)
            max_trade_duration = np.nanmax(trade_summary.trade_duration.values)
            max_drawdown = min(account_history.drawdown)
            total_fees = trade_summary.fees.sum()
            
            starting_balance = account_history.balance[0]
            ending_balance = account_history.balance[-1]
            ending_NAV = account_history.NAV[-1]
            abs_return = ending_balance - starting_balance
            pc_return = 100 * abs_return / starting_balance
            
            backtest_results['all_trades']['starting_balance'] = starting_balance
            backtest_results['all_trades']['ending_balance'] = ending_balance
            backtest_results['all_trades']['ending_NAV']    = ending_NAV
            backtest_results['all_trades']['abs_return']    = abs_return
            backtest_results['all_trades']['pc_return']     = pc_return
            backtest_results['all_trades']['avg_win']       = avg_win
            backtest_results['all_trades']['max_win']       = max_win
            backtest_results['all_trades']['avg_loss']      = avg_loss
            backtest_results['all_trades']['max_loss']      = max_loss
            backtest_results['all_trades']['win_rate']      = win_rate
            backtest_results['all_trades']['win_streak']    = longest_win_streak
            backtest_results['all_trades']['lose_streak']   = longest_lose_streak
            backtest_results['all_trades']['longest_trade'] = str(timedelta(seconds = int(max_trade_duration)))
            backtest_results['all_trades']['shortest_trade'] = str(timedelta(seconds = int(min_trade_duration)))
            backtest_results['all_trades']['avg_trade_duration'] = str(timedelta(seconds = int(avg_trade_duration)))
            backtest_results['all_trades']['net_pl']        = cpl.values[-1]
            backtest_results['all_trades']['max_drawdown']  = max_drawdown
            backtest_results['all_trades']['total_fees']    = total_fees
            
        # Cancelled and open orders
        cancelled_orders = self._broker.get_orders(instrument, 'cancelled')
        open_trades = self._broker.get_positions(instrument)
        backtest_results['no_open'] = len(open_trades)
        backtest_results['no_cancelled'] = len(cancelled_orders)
        
        # Long trades
        long_trades = trade_summary[trade_summary['direction'] > 0]
        no_long = len(long_trades)
        backtest_results['long_trades'] = {}
        backtest_results['long_trades']['no_trades'] = no_long
        if no_long > 0:
            long_wins       = long_trades[long_trades.profit > 0]
            avg_long_win    = np.mean(long_wins.profit)
            max_long_win    = np.max(long_wins.profit)
            long_loss       = long_trades[long_trades.profit < 0]
            avg_long_loss   = abs(np.mean(long_loss.profit))
            max_long_loss   = abs(np.min(long_loss.profit))
            long_wr         = 100*len(long_trades[long_trades.profit > 0])/no_long
            
            backtest_results['long_trades']['avg_long_win']     = avg_long_win
            backtest_results['long_trades']['max_long_win']     = max_long_win 
            backtest_results['long_trades']['avg_long_loss']    = avg_long_loss
            backtest_results['long_trades']['max_long_loss']    = max_long_loss
            backtest_results['long_trades']['long_wr']          = long_wr
            
        # Short trades
        short_trades    = trade_summary[trade_summary['direction'] < 0]
        no_short        = len(short_trades)
        backtest_results['short_trades'] = {}
        backtest_results['short_trades']['no_trades'] = no_short
        if no_short > 0:
            short_wins      = short_trades[short_trades.profit > 0]
            avg_short_win   = np.mean(short_wins.profit)
            max_short_win   = np.max(short_wins.profit)
            short_loss      = short_trades[short_trades.profit < 0]
            avg_short_loss  = abs(np.mean(short_loss.profit))
            max_short_loss  = abs(np.min(short_loss.profit))
            short_wr        = 100*len(short_trades[short_trades.profit > 0])/no_short
            
            backtest_results['short_trades']['avg_short_win']   = avg_short_win
            backtest_results['short_trades']['max_short_win']   = max_short_win
            backtest_results['short_trades']['avg_short_loss']  = avg_short_loss
            backtest_results['short_trades']['max_short_loss']  = max_short_loss
            backtest_results['short_trades']['short_wr']        = short_wr
        
        return backtest_results
    
    
    def multibot_backtest_analysis(self, bots: list = None) -> dict:
        """Analyses backtest results of multiple bots to create an overall 
        performance summary.
        
        Parameters
        -----------
        bots : list[AutoTraderBot]
            A list of AutoTrader bots to analyse.
        
        Returns
        -------
        backtest_results : dict
            A dictionary of backtest results.
        
        Notes
        -----
        If no bots are supplied, the backtest will be analysed for all bots 
        assiged during the backtest.
        
        See Also
        --------
        get_bots_deployed
        """
        
        instruments = []
        win_rate    = []
        no_trades   = []
        avg_win     = []
        max_win     = []
        avg_loss    = []
        max_loss    = []
        no_long     = []
        no_short    = []
        
        if bots is None:
            bots = self._bots_deployed
        
        for bot in bots:
            backtest_results = self.analyse_backtest(bot)
            
            instruments.append(bot.instrument)
            no_trades.append(backtest_results['no_trades'])
            if backtest_results['no_trades'] > 0:
                win_rate.append(backtest_results['all_trades']['win_rate'])
                avg_win.append(backtest_results['all_trades']['avg_win'])
                max_win.append(backtest_results['all_trades']['max_win'])
                avg_loss.append(backtest_results['all_trades']['avg_loss'])
                max_loss.append(backtest_results['all_trades']['max_loss'])
                no_long.append(backtest_results['long_trades']['no_trades'])
                no_short.append(backtest_results['short_trades']['no_trades'])
            else:
                win_rate.append(np.nan)
                avg_win.append(np.nan)
                max_win.append(np.nan)
                avg_loss.append(np.nan)
                max_loss.append(np.nan)
                no_long.append(np.nan)
                no_short.append(np.nan)
        
        multibot_backtest_results = pd.DataFrame(data={'win_rate': win_rate,
                                                       'no_trades': no_trades,
                                                       'avg_win': avg_win,
                                                       'max_win': max_win,
                                                       'avg_loss': avg_loss,
                                                       'max_loss': max_loss,
                                                       'no_long': no_long,
                                                       'no_short': no_short},
                                                 index=instruments)
        
        return multibot_backtest_results
        
    
    def print_multibot_backtest_results(self, backtest_results: dict = None) -> None:
        """Prints to console the backtest results of a multi-bot backtest.
        
        Parameters
        -----------
        backtest_results : dict
            A dictionary containing backtest results.
            
        See Also
        --------
        Analyse backtest method to generate backtest_results.
        """
        
        bot = self._bots_deployed[0]
        account_history = bot.backtest_summary['account_history']
        
        start_date = account_history.index[0]
        end_date = account_history.index[-1]
        
        starting_balance = account_history.balance[0]
        ending_balance = account_history.balance[-1]
        ending_NAV = account_history.NAV[-1]
        abs_return = ending_balance - starting_balance
        pc_return = 100 * abs_return / starting_balance
        
        print("\n---------------------------------------------------")
        print("            MultiBot Backtest Results")
        print("---------------------------------------------------")
        print("Start date:              {}".format(start_date))
        print("End date:                {}".format(end_date))
        print("Starting balance:        ${}".format(round(starting_balance, 2)))
        print("Ending balance:          ${}".format(round(ending_balance, 2)))
        print("Ending NAV:              ${}".format(round(ending_NAV, 2)))
        print("Total return:            ${} ({}%)".format(round(abs_return, 2), 
                                          round(pc_return, 1)))
        
        print("Instruments traded: ", backtest_results.index.values)
        print("Total no. trades:   ", backtest_results.no_trades.sum())
        print("Short trades:       ", backtest_results.no_short.sum(),
              "({}%)".format(round(100*backtest_results.no_short.sum()/backtest_results.no_trades.sum(),2)))
        print("Long trades:        ", backtest_results.no_long.sum(),
              "({}%)".format(round(100*backtest_results.no_long.sum()/backtest_results.no_trades.sum(),2)))
        
        print("\nInstrument win rates (%):")
        print(backtest_results[['win_rate', 'no_trades']])
        print("\nMaximum/Average Win/Loss breakdown ($):")
        print(backtest_results[["max_win", "max_loss", "avg_win", "avg_loss"]])
        print("\nAverage Reward to Risk Ratio:")
        print(round(backtest_results.avg_win / backtest_results.avg_loss,1))
        print("")
        
    
    @staticmethod
    def print_backtest_results(backtest_results: dict) -> None:
        """Prints backtest results.

        Parameters
        ----------
        backtest_results : dict
            The backtest results dictionary.

        Returns
        -------
        None
            Backtest results will be printed.

        See Also
        ----------
        analyse_backtest
        """
        start_date = backtest_results['start'].strftime("%b %d %Y %H:%M:%S")
        end_date = backtest_results['end'].strftime("%b %d %Y %H:%M:%S")
        
        no_trades   = backtest_results['no_trades']
        if no_trades > 0:
            win_rate    = backtest_results['all_trades']['win_rate']
            abs_return  = backtest_results['all_trades']['abs_return']
            pc_return   = backtest_results['all_trades']['pc_return']
            max_drawdown = backtest_results['all_trades']['max_drawdown']
            max_win     = backtest_results['all_trades']['max_win']
            avg_win     = backtest_results['all_trades']['avg_win']
            max_loss    = backtest_results['all_trades']['max_loss']
            avg_loss    = backtest_results['all_trades']['avg_loss']
            longest_win_streak = backtest_results['all_trades']['win_streak']
            longest_lose_streak = backtest_results['all_trades']['lose_streak']
            total_fees = backtest_results['all_trades']['total_fees']
            starting_balance = backtest_results['all_trades']['starting_balance']
            ending_balance = backtest_results['all_trades']['ending_balance']
            ending_NAV = backtest_results['all_trades']['ending_NAV']
        
        print("\n----------------------------------------------")
        print("               Backtest Results")
        print("----------------------------------------------")
        if no_trades > 0:
            print("Start date:              {}".format(start_date))
            print("End date:                {}".format(end_date))
            print("Starting balance:        ${}".format(round(starting_balance, 2)))
            print("Ending balance:          ${}".format(round(ending_balance, 2)))
            print("Ending NAV:              ${}".format(round(ending_NAV, 2)))
            print("Total return:            ${} ({}%)".format(round(abs_return, 2), 
                                              round(pc_return, 1)))
            print("Total no. trades:        {}".format(no_trades))
            print("Total fees:              ${}".format(round(total_fees, 3)))
            print("Backtest win rate:       {}%".format(round(win_rate, 1)))
            print("Maximum drawdown:        {}%".format(round(max_drawdown*100, 2)))
            print("Max win:                 ${}".format(round(max_win, 2)))
            print("Average win:             ${}".format(round(avg_win, 2)))
            print("Max loss:                -${}".format(round(max_loss, 2)))
            print("Average loss:            -${}".format(round(avg_loss, 2)))
            print("Longest win streak:      {} trades".format(longest_win_streak))
            print("Longest losing streak:   {} trades".format(longest_lose_streak))
            print("Average trade duration:  {}".format(backtest_results['all_trades']['avg_trade_duration']))
            
        else:
            print("No trades taken.")
        
        no_open = backtest_results['no_open']
        no_cancelled = backtest_results['no_cancelled']
        
        if no_open > 0:
            print("Orders still open:       {}".format(no_open))
        if no_cancelled > 0:
            print("Cancelled orders:        {}".format(no_cancelled))
        
        # Long trades
        no_long = backtest_results['long_trades']['no_trades']
        print("\n            Summary of long trades")
        print("----------------------------------------------")
        if no_long > 0:
            avg_long_win = backtest_results['long_trades']['avg_long_win']
            max_long_win = backtest_results['long_trades']['max_long_win']
            avg_long_loss = backtest_results['long_trades']['avg_long_loss']
            max_long_loss = backtest_results['long_trades']['max_long_loss']
            long_wr = backtest_results['long_trades']['long_wr']
            
            print("Number of long trades:   {}".format(no_long))
            print("Long win rate:           {}%".format(round(long_wr, 1)))
            print("Max win:                 ${}".format(round(max_long_win, 2)))
            print("Average win:             ${}".format(round(avg_long_win, 2)))
            print("Max loss:                -${}".format(round(max_long_loss, 2)))
            print("Average loss:            -${}".format(round(avg_long_loss, 2)))
        else:
            print("There were no long trades.")
          
        # Short trades
        no_short = backtest_results['short_trades']['no_trades']
        print("\n             Summary of short trades")
        print("----------------------------------------------")
        if no_short > 0:
            avg_short_win = backtest_results['short_trades']['avg_short_win']
            max_short_win = backtest_results['short_trades']['max_short_win']
            avg_short_loss = backtest_results['short_trades']['avg_short_loss']
            max_short_loss = backtest_results['short_trades']['max_short_loss']
            short_wr = backtest_results['short_trades']['short_wr']
            
            print("Number of short trades:  {}".format(no_short))
            print("short win rate:          {}%".format(round(short_wr, 1)))
            print("Max win:                 ${}".format(round(max_short_win, 2)))
            print("Average win:             ${}".format(round(avg_short_win, 2)))
            print("Max loss:                -${}".format(round(max_short_loss, 2)))
            print("Average loss:            -${}".format(round(avg_short_loss, 2)))
            
        else:
            print("There were no short trades.")

    
    def _main(self) -> None:
        """Run AutoTrader with configured settings.
        """
        # Load configuration
        if self._global_config_dict is not None:
            # Use global config dict provided
            global_config = self._global_config_dict
        else:
            # Try load from file
            global_config_fp = os.path.join(self._home_dir, 'config', 'GLOBAL.yaml')
            if os.path.isfile(global_config_fp):
                global_config = read_yaml(global_config_fp)
            else:
                global_config = None
        
        # Check feed
        if global_config is None and self._feed.lower() in ['oanda']:
            raise Exception(f'Data feed "{self._feed}" requires global '+ \
                            'configuration. If a config file already '+ \
                            'exists, make sure to specify the home_dir.')
            
        broker_config = get_config(self._environment, global_config, self._feed)
        
        if self._account_id is not None:
            # Overwrite default account in global config
            broker_config['ACCOUNT_ID'] = self._account_id
        
        # Append broker verbosity to broker_config
        broker_config['verbosity'] = self._broker_verbosity
        
        self._assign_broker(broker_config)
        self._configure_emailing(global_config)
        
        if self._backtest_mode:
            NAV     = []
            balance = []
            margin  = []
            tradetimes = []
        
        if int(self._verbosity) > 0:
            if self._backtest_mode:
                print("Beginning new backtest.")

            elif self._scan_mode:
                print("AutoTrader - AutoScan")
                print("Time: {}\n".format(datetime.now().strftime("%A, %B %d %Y, "+
                                                                  "%H:%M:%S")))
            else:
                print("AutoTrader Livetrade")
                print("--------------------")
                print("Current time: {}".format(datetime.now().strftime("%A, %B %d %Y, "+
                                                                  "%H:%M:%S")))
        
        # Assign strategy to bot for each instrument in watchlist 
        for strategy in self._strategy_configs:
            for instrument in self._strategy_configs[strategy]['WATCHLIST']:
                data_dict = self._local_data[instrument] \
                    if self._local_data is not None else None
                quote_data_path = self._local_quote_data[instrument] \
                    if self._local_quote_data is not None else None
                auxdata = self._auxdata[instrument] \
                    if self._auxdata is not None else None
                
                strategy_class = self._strategy_configs[strategy]['CLASS']
                strategy_dict = {'config': self._strategy_configs[strategy],
                                 'class': self._strategy_classes[strategy_class] \
                                     if strategy_class in self._strategy_classes else None,
                                 'shutdown_method': self._shutdown_methods[strategy]}
                bot = AutoTraderBot(instrument, strategy_dict,
                                    self._broker, self._data_start, data_dict, 
                                    quote_data_path, auxdata, self)
                self._bots_deployed.append(bot)
                
        if int(self._verbosity) > 0 and self._backtest_mode:
            print("\nTrading...\n")
            backtest_start_time = timeit.default_timer()
            
        # Begin trading
        if self._run_mode.lower() == 'continuous':
            # Running in continuous update mode
            if self._backtest_mode:
                # Backtesting
                end_time = self._data_end # datetime
                timestamp = self._data_start # datetime
                while timestamp <= end_time:
                    # Update each bot with latest data to generate signal
                    for bot in self._bots_deployed:
                        bot._update(timestamp=timestamp)
                        
                    # Update backtest tracking stats
                    NAV.append(self._broker.NAV)
                    balance.append(self._broker.portfolio_balance)
                    margin.append(self._broker.margin_available)
                    tradetimes.append(timestamp)
                    
                    # Iterate through time
                    timestamp += self._timestep
        
            else:
                # Live trading
                instance_id = self._get_instance_id()
                instance_str = f"autotrader_instance_{instance_id}" if \
                    self._instance_str is None else self._instance_str
                instance_file_exists = self._check_instance_file(instance_str, True)
                deploy_time = time.time()
                while instance_file_exists:
                    for bot in self._bots_deployed:
                        bot._update(timestamp=datetime.now(timezone.utc))
                    time.sleep(self._timestep.seconds - ((time.time() - \
                                deploy_time) % self._timestep.seconds))
                    instance_file_exists = self._check_instance_file(instance_str)
                    
        elif self._run_mode.lower() == 'periodic':
            # Trading in periodic update mode
            if self._backtest_mode:
                # Backtesting
                self._check_bot_data()
                start_range, end_range = self._bots_deployed[0]._get_iteration_range()
                for i in range(start_range, end_range):
                    # Update each bot with latest data to generate signal
                    for bot in self._bots_deployed:
                        bot._update(i=i)
                        
                    # Update backtest tracking
                    NAV.append(self._broker.NAV)
                    balance.append(self._broker.portfolio_balance)
                    margin.append(self._broker.margin_available)
                    tradetimes.append(self._bots_deployed[0].data.index[i])
        
            else:
                # Live trading
                bot._update(i=-1) # Process most recent signal
        
        if int(self._verbosity) > 0 and self._backtest_mode:
            backtest_end_time = timeit.default_timer()
        
        # Run strategy-specific shutdown routines
        for bot in self._bots_deployed:
            bot._strategy_shutdown()
            
        # Run instance shut-down routine
        if self._backtest_mode:
            # Create backtest summary for each bot 
            for bot in self._bots_deployed:
                bot._create_backtest_summary(balance, NAV, margin, tradetimes)            
            
            if int(self._verbosity) > 0:
                print(f"Backtest complete (runtime {round((backtest_end_time - backtest_start_time), 3)} s).")
                if len(self._bots_deployed) == 1:
                    bot = self._bots_deployed[0]
                    backtest_results = self.analyse_backtest(bot)
                    self.print_backtest_results(backtest_results)
                    
                else:
                    self.multibot_backtest_results = self.multibot_backtest_analysis()
                    self.print_multibot_backtest_results(self.multibot_backtest_results)
                    
                    print("Results for multiple-instrument backtests have been")
                    print("written to AutoTrader.multibot_backtest_results.")
                    print("Individual bot results can be found in the")
                    print("'bots_deployed' attribute of the AutoTrader instance.")
            
            if self._show_plot:
                if len(self._bots_deployed) == 1:
                    if len(self._broker.trades) > 0:
                        self.plot_backtest(bot=self._bots_deployed[0])
                
                else:
                    # Backtest run with multiple bots
                    self.plot_multibot_backtest()
        
        elif self._scan_mode and self._show_plot:
            # Show plots for scanned instruments
            for bot in self._bots_deployed:
                ap = self._instantiate_autoplot(bot.data)
                ap.plot(indicators = bot.strategy.indicators, 
                        instrument = bot.instrument)
                time.sleep(0.3)
        
        else:
            # Live trade complete, run livetrade specific shutdown routines
            if self._broker_name.lower() == 'ib':
                self._broker._disconnect()
            
            elif self._virtual_livetrading:
                # TODO - write broker stats to file (trade summary and so on)
                # look to AutoTraderBot._create_backtest_summary for process
                pass
        

    def _clear_strategies(self) -> None:
        """Removes all strategies saved in autotrader instance.
        """
        self._strategy_configs = {}
    
    
    def _clear_bots(self) -> None:
        """Removes all deployed bots in autotrader instance.
        """
        self._bots_deployed = []
    
    
    def _instantiate_autoplot(self, data: pd.DataFrame) -> AutoPlot:
        """Creates instance of AutoPlot.

        Parameters
        ----------
        data : pd.DataFrame
            The data to instantiate AutoPlot with.

        Raises
        ------
        Exception
            When attempting to plot on missing data timeframe.

        Returns
        -------
        AutoPlot
            An instance of AutoPlot.
        """
        if self._chart_timeframe == 'default':
            ap = AutoPlot(data)
        else:
            # Instantiate AutoPlot with requested chart timeframe
            if self._chart_timeframe in self._bots_deployed[0].MTF_data.keys():
                # Valid timeframe requested
                ap = AutoPlot(self._bots_deployed[0].MTF_data[self._chart_timeframe])
                ap._add_backtest_price_data(data) # provide nominal timeframe data for merge operations
            else:
                warning_str = f'The chart timeframe requested ({self._chart_timeframe}) was not found ' + \
                    'in the MTF data. Please ensure that the timeframe provided matches ' + \
                    'the format provided in the strategy configuration file, or the local ' + \
                    'data provided.'
                raise Exception(warning_str)
                
        # Assign attributes
        ap.configure(max_indis_over=self._max_indis_over, 
                     max_indis_below = self._max_indis_below,
                     fig_tools = self._fig_tools, 
                     ohlc_height = self._ohlc_height, 
                     ohlc_width = self._ohlc_width, 
                     top_fig_height = self._top_fig_height,
                     bottom_fig_height = self._bottom_fig_height,
                     jupyter_notebook = self._jupyter_notebook,
                     show_cancelled = self._show_cancelled,
                     chart_theme = self._chart_theme,
                     use_strat_plot_data = self._use_strat_plot_data,)
        
        return ap
    
    
    def _update_strategy_watchlist(self) -> None:
        """Updates the watchlist of each strategy with the scan watchlist.
        """
        for strategy in self._strategy_configs:
            self._strategy_configs[strategy]['WATCHLIST'] = self._scan_watchlist
    
    
    def _assign_broker(self, broker_config: dict) -> None:
        """Configures and assigns appropriate broker for trading.
        """
        # Import relevant broker and utilities modules
        broker_module = importlib.import_module(f'autotrader.brokers.{self._broker_name}.broker')
        utils_module = importlib.import_module(f'autotrader.brokers.{self._broker_name}.utils')
        
        # Create broker and utils instances
        utils = utils_module.Utils()
        broker = broker_module.Broker(broker_config, utils)
        
        if self._backtest_mode or self._virtual_livetrading:
            # Using virtual broker, initialise account
            if int(self._verbosity) > 0 and self._backtest_mode:
                banner = pyfiglet.figlet_format("AutoBacktest")
                print(banner)
            broker._make_deposit(self._virtual_initial_balance)
            broker.fee = self._virtual_spread
            broker.leverage = self._virtual_leverage
            broker.commission = self._virtual_commission
            broker.spread = self._virtual_spread
            broker.base_currency = self._base_currency
            broker.hedging = self._virtual_broker_hedging
            broker.margin_closeout = self._virtual_margin_call
        
        self._broker = broker
        self._broker_utils = utils
    
    
    def _configure_emailing(self, global_config: dict) -> None:
        """Configure email settings.
        """
        if int(self._notify) > 0:
            host_email = None
            mailing_list = None
            
            if "EMAILING" in global_config:
                # Look for host email and mailing list in strategy config, if it
                # was not picked up in strategy config
                if "MAILING_LIST" in global_config["EMAILING"] and mailing_list is None:
                    mailing_list    = global_config["EMAILING"]["MAILING_LIST"]
                if "HOST_ACCOUNT" in global_config["EMAILING"] and host_email is None:
                    host_email      = global_config["EMAILING"]["HOST_ACCOUNT"]
            
            if host_email is None:
                print("Warning: email host account not provided.")
            if mailing_list is None:
                print("Warning: no mailing list provided.")
                
            email_params = {'mailing_list': mailing_list,
                            'host_email': host_email}
            self._email_params = email_params
            
            logfiles_path = os.path.join(self._home_dir, 'logfiles')
            order_summary_fp = os.path.join(logfiles_path, 'order_history.txt')
            
            if not os.path.isdir(logfiles_path):
                os.mkdir(logfiles_path)
            
            self._order_summary_fp = order_summary_fp


    def _run_optimise(self) -> None:
        """Runs optimisation of strategy parameters.
        """
        
        # Modify verbosity for optimisation
        verbosity = self._verbosity
        self._verbosity = 0
        self._show_plot = False
        
        self.objective      = 'profit + MDD'
        
        ''' --------------------------------------------------------------- '''
        '''                          Unpack user options                    '''
        ''' --------------------------------------------------------------- '''
        
        # Look in self._strategy_configs for config
        if len(self._strategy_configs) > 1:
            print("Error: please optimise one strategy at a time.")
            print("Exiting.")
            sys.exit(0)
        else:
            config_dict = self._strategy_configs[list(self._strategy_configs.keys())[0]]
                
        ''' --------------------------------------------------------------- '''
        '''                      Define optimisation inputs                 '''
        ''' --------------------------------------------------------------- '''
        my_args = (config_dict, self._opt_params, self._verbosity)
        
        ''' --------------------------------------------------------------- '''
        '''                             Run Optimiser                       '''
        ''' --------------------------------------------------------------- '''
        start = timeit.default_timer()
        result = brute(func         = self._optimisation_helper_function, 
                       ranges       = self._bounds, 
                       args         = my_args, 
                       Ns           = self._Ns,
                       full_output  = True)
        stop = timeit.default_timer()
        
        ''' --------------------------------------------------------------- '''
        '''      Delete historical data file after running optimisation     '''
        ''' --------------------------------------------------------------- '''
        granularity             = config_dict["INTERVAL"]
        pair                    = config_dict["WATCHLIST"][0]
        historical_data_name    = 'hist_{0}{1}.csv'.format(granularity, pair)
        historical_quote_data_name = 'hist_{0}{1}_quote.csv'.format(granularity, pair)
        historical_data_file_path = os.path.join(self._home_dir, 
                                                 'price_data',
                                                 historical_data_name)
        historical_quote_data_file_path = os.path.join(self._home_dir, 
                                                       'price_data',
                                                       historical_quote_data_name)
        os.remove(historical_data_file_path)
        os.remove(historical_quote_data_file_path)
        
        opt_params = result[0]
        opt_value = result[1]
        
        ''' --------------------------------------------------------------- '''
        '''                           Print output                          '''
        ''' --------------------------------------------------------------- '''
        print("\nOptimisation complete.")
        print('Time to run: {}s'.format(round((stop - start), 3)))
        print("Optimal parameters:")
        print(opt_params)
        print("Objective:")
        print(opt_value)
        
        # Reset verbosity
        self._verbosity = verbosity
    
    
    def _optimisation_helper_function(self, params: list, config_dict: dict, 
                                      opt_params: list, verbosity: int) -> float:
        """Helper function for optimising strategy parameters in AutoTrader.
        This function will parse the ordered params into the config dict.
        """
        
        ''' ------------------------------------------------------------------ '''
        '''   Edit strategy parameters in config_dict using supplied params    '''
        ''' ------------------------------------------------------------------ '''
        for parameter in config_dict['PARAMETERS']:
            if parameter in opt_params:
                config_dict['PARAMETERS'][parameter] = params[opt_params.index(parameter)]
            else:
                continue
        
        ''' ------------------------------------------------------------------ '''
        '''           Run AutoTrader and evaluate objective function           '''
        ''' ------------------------------------------------------------------ '''
        self._clear_strategies()
        self._clear_bots()
        self.add_strategy(config_dict = config_dict)
        self._main()
        
        bot = self._bots_deployed[0]
            
        try:
            backtest_results = self.analyse_backtest(bot)
            objective = -backtest_results['all_trades']['ending_NAV']
        except:
            objective = 1000
                              
        print("Parameters/objective:", params, "/", round(objective,3))
        
        return objective
    
    
    def _check_bot_data(self) -> None:
        """Function to compare lengths of bot data. 
        """
        
        data_lengths = [len(bot.data) for bot in self._bots_deployed]
        
        if min(data_lengths) != np.mean(data_lengths):
            print("\nWarning: mismatched data lengths detected. Correcting via row reduction.")
            self._normalise_bot_data()
    
    
    def _normalise_bot_data(self) -> None:
        """Function to normalise the data of mutliple bots so that their
        indexes are equal, allowing backtesting.
        """
        
        # Construct list of bot data
        data = [bot.data for bot in self._bots_deployed]
        
        for i, dat in enumerate(data):
            # Initialise common index
            comm_index = dat.index
            
            # Update common index by intersection with other data 
            for j, dat_2 in enumerate(data):
                comm_index = comm_index.intersection(dat_2.index)
            
            # Adjust bot data using common indexes
            adj_data = dat[dat.index.isin(comm_index)]
            
            # Re-assign bot data
            self._bots_deployed[i]._replace_data(adj_data)
    
    
    def _get_instance_id(self):
        """Returns an ID for the active AutoTrader instance.
        """
        dirpath = os.path.join(self._home_dir, 'active_bots')
        
        # Check if active_bots directory exists
        if not os.path.isdir(dirpath):
            # Directory does not exist, create it
            os.mkdir(dirpath)
            instance_id = 1
            
        else:
            # Directory exists, find highest instance
            instances = [f for f in os.listdir(dirpath) if \
                         os.path.isfile(os.path.join(dirpath, f))]
            
            last_id = 0
            for instance in instances:
                if 'autotrader_instance_' in instance:
                    # Ignore custom instance strings
                    last_id = int(instance.split('_')[-1])
            
            instance_id = last_id + 1
        
        return instance_id
    
    
    def _check_instance_file(self, instance_str, initialisation=False):
        """Checks if the AutoTrader instance exists.
        """
        if initialisation:
            # Create the file
            filepath = os.path.join(self._home_dir, 'active_bots', instance_str)
            with open(filepath, mode='w') as f:
                f.write("This instance of AutoTrader contains the following bots:\n")
                for bot in self._bots_deployed:
                    f.write(bot._strategy_name + f" ({bot.instrument})\n")
            instance_file_exists = True
            
            if int(self._verbosity) > 0:
                print(f"Active AutoTrader instance file: active_bots/{instance_str}")
        
        else:
            dirpath = os.path.join(self._home_dir, 'active_bots')
            instances = [f for f in os.listdir(dirpath) if \
                         os.path.isfile(os.path.join('active_bots', f))]
            instance_file_exists = instance_str in instances
        
        if int(self._verbosity) > 0 and not instance_file_exists:
            print(f"Instance file '{instance_str}' deleted. AutoTrader",
                  "will now shut down.")
        
        return instance_file_exists
        



import random
import ib_insync
import numpy as np
from autotrader.brokers.trading import Order, Trade, Position
from autotrader.brokers.ib.utils import Utils


class Broker:
    """AutoTrader-InteractiveBrokers API interface.
    
    Attributes
    ----------
    utils : Utils
        The broker utilities.
    ib : ib_insync connection
        Used to query IB.
    account : str
        The active IB account.
    
    Notes
    -----
        - when closing a position using close_position(), if there are attached SL
          and/or TP orders, they must be closed manually using cancel_pending_order().
          Usually only one of the pair needs to be cancelled, and the other will too.
        - required signal_dict keys for different security types (eg. futures 
          require symbol, exchange and contract_month)
        - The products localSymbol will always take precedence over the symbol.
          As such, it should be used as much as possible to avoid accidental 
          actions.
    """
    
    def __init__(self, config: dict, utils: Utils = None) -> None:
        """Initialise AutoTrader-Interactive Brokers API interface.
        
        Parameters
        ----------
        config : dict
            The IB configuration dictionary. This can contain the host, port, 
            clientID and read_only boolean flag.
        utils : Utils, optional
            Broker utilities class instance. The default is None.
        """
        self.utils = utils if utils is not None else Utils()
        
        self.host = config['host'] if 'host' in config else '127.0.0.1'
        self.port = config['port'] if 'port' in config else 7497
        self.client_id = config['clientID'] if 'clientID' in config else random.randint(0, 9999)
        self.read_only = config['read_only'] if 'read_only' in config else False
        self.account = config['account'] if 'account' in config else ''
        
        self.ib = ib_insync.IB()
        self._check_connection()
        self.account = self._get_account() if self.account == '' else self.account
        
        
    def __repr__(self):
        return 'AutoTrader-InteractiveBrokers interface'
    
    
    def __str__(self):
        return 'AutoTrader-InteractiveBrokers interface'
    
    
    def get_NAV(self) -> float:
        """Returns the net asset/liquidation value of the account.
        """
        self._check_connection()
        summary = self.get_summary()
        return float(summary['NetLiquidation']['value'])
    
    
    def get_balance(self) -> float:
        """Returns account balance.
        """
        self._check_connection()
        summary = self.get_summary()
        return float(summary['TotalCashValue']['value'])
        
    
    def place_order(self, order: Order, **kwargs) -> None:
        """Disassembles order_details dictionary to place order.

        Parameters
        ----------
        order: Order
            The AutoTrader Order.

        Returns
        -------
        None
            Orders will be submitted to IB.
        """
        self._check_connection()
        
        # Call order to set order time
        order()
        
        if order.order_type == 'market':
            self._place_market_order(order)
        elif order.order_type == 'stop-limit':
            self._place_stop_limit_order(order)
        elif order.order_type == 'limit':
            self._place_limit_order(order)
        elif order.order_type == 'close':
            self._close_position(order)
        else:
            print("Order type not recognised.")
        
        self._refresh()
        
    
    def get_orders(self, instrument: str = None, **kwargs) -> dict:
        """Returns all pending orders (have not been filled) in the account.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument's symbol. The default is None.

        Returns
        -------
        dict
            Pending orders for the requested instrument. If no instrument is provided,
            all pending orders will be returned.
        """
        self._check_connection()
        
        # Get all open trades
        open_trades = self.ib.openTrades()
        
        pending_orders = {}
        for trade in open_trades:
            trade_dict = trade.dict()
            contract = trade_dict['contract']
            order_dict = trade_dict['order'].dict()
            order_status_dict = trade_dict['orderStatus'].dict()
            order_status = order_status_dict['status']
            
            if order_status in ib_insync.OrderStatus.ActiveStates:
                # Order is still active (not yet filled)
                new_order = {}
                new_order['order_ID']           = order_dict['orderId']
                new_order['order_type']         = order_dict['orderType']
                new_order['order_stop_price']   = order_dict['auxPrice']
                new_order['order_limit_price']  = order_dict['lmtPrice']
                new_order['direction']          = 1 if order_dict['action'] == 'BUY' else -1
                new_order['order_time']         = None
                new_order['instrument']         = contract.symbol
                new_order['size']               = order_dict['totalQuantity']
                new_order['order_price']        = None
                new_order['take_profit']        = None
                new_order['take_distance']      = None
                new_order['stop_type']          = None
                new_order['stop_distance']      = None
                new_order['stop_loss']          = None
                new_order['related_orders']     = None
                new_order['granularity']        = None
                new_order['strategy']           = None
                
                if instrument is not None and contract.symbol == instrument:
                    pending_orders[new_order['order_ID']] = Order._from_dict(new_order)
                elif instrument is None:
                    pending_orders[new_order['order_ID']] = Order._from_dict(new_order)
            
        return pending_orders
    
    
    def cancel_order(self, order_id: int, **kwargs) -> list:
        """Cancels pending order by order ID.
        
        Parameters
        ----------
        order_id : int
            The ID of the order to be concelled.

        Returns
        -------
        list
            A list of the cancelled trades.

        """
        self._check_connection()
        
        open_trades = self.ib.openTrades()
        cancelled_trades = []
        for trade in open_trades:
            order = trade.order
            if order.orderId == order_id:
                cancel_trade = self.ib.cancelOrder(order)
                cancelled_trades.append(cancel_trade)
        
        return cancelled_trades
    
    
    def get_trades(self, instrument: str = None, **kwargs) -> dict:
        """Returns the open trades held by the account. 

        Parameters
        ----------
        instrument : str, optional
            The trading instrument's symbol. The default is None.

        Returns
        -------
        dict
            The open trades.
        """
        self._check_connection()
        
        # Get all open trades
        all_open_trades = self.ib.openTrades()
        
        open_trades = {}
        for trade in all_open_trades:
            trade_dict = trade.dict()
            contract = trade_dict['contract']
            order_dict = trade_dict['order'].dict()
            order_status_dict = trade_dict['orderStatus'].dict()
            order_status = order_status_dict['status']
            
            if order_status == 'Filled':
                # Trade order has been filled
                new_trade = {}
                new_trade['order_ID']           = order_dict['orderId']
                new_trade['order_stop_price']   = order_dict['auxPrice']
                new_trade['order_limit_price']  = order_dict['lmtPrice']
                new_trade['direction']          = 1 if order_dict['action'] == 'BUY' else -1
                new_trade['order_time']         = None
                new_trade['instrument']         = contract.symbol
                new_trade['size']               = order_dict['totalQuantity']
                new_trade['order_price']        = None
                new_trade['entry_price']        = order_status_dict['lastFillPrice']
                new_trade['order_type']         = None
                new_trade['take_profit']        = None
                new_trade['take_distance']      = None
                new_trade['stop_type']          = None
                new_trade['stop_distance']      = None
                new_trade['stop_loss']          = None
                new_trade['related_orders']     = None
                new_trade['granularity']        = None
                new_trade['strategy']           = None
                
                if instrument is not None and contract.symbol == instrument:
                    open_trades[new_trade['order_ID']] = Trade(new_trade)
                elif instrument is None:
                    open_trades[new_trade['order_ID']] = Trade(new_trade)
        
        return open_trades
    
    
    def get_trade_details(self, trade_ID: str, **kwargs) -> dict:
        """Returns the details of the trade specified by trade_ID.

        Parameters
        ----------
        trade_ID : str
            The ID of the trade.

        Returns
        -------
        dict
            The details of the trade.
        """
        raise NotImplementedError("This method is not available.")
        # self._check_connection()
        # TODO - implement (?)
        # return {}
    
    
    def get_positions(self, instrument: str = None, **kwargs) -> dict:
        """Gets the current positions open on the account.
        
        Parameters
        ----------
        instrument : str, optional
            The trading instrument's symbol. This can be either the naive
            symbol, or the localSymbol. The default is None.
            
        Returns
        -------
        open_positions : dict
            A dictionary containing details of the open positions.
        
        Notes
        -----
        This function returns the position in an underlying product. If 
        there are multiple contracts on the underlying, they will be 
        appended to the returned Position objects' "contracts" and 
        "portfolio_items" attributes.
        """
        
        def adjust_position(existing_pos, portfolio_item, pos_dict):
            # Position item already exists, append portfolio item
            existing_pos.long_units += pos_dict['long_units']
            existing_pos.long_PL += pos_dict['long_PL']
            existing_pos.short_units += pos_dict['short_units']
            existing_pos.short_PL += pos_dict['short_PL']
            existing_pos.net_position += pos_dict['net_position']
            existing_pos.PL += pos_dict['PL']
            existing_pos.contracts.append(portfolio_item.contract)
            existing_pos.portfolio_items.append(portfolio_item)
        
        self._check_connection()
        
        all_portfolio_items = self.ib.portfolio()
        open_positions = {}
        for item in all_portfolio_items:
            units = item.position
            pnl = item.unrealizedPNL
            pos_symbol = item.contract.symbol
            pos_dict = {'long_units': units if np.sign(units) > 0 else 0,
                        'long_PL': pnl if np.sign(units) > 0 else 0,
                        'short_units': abs(units) if np.sign(units) < 0 else 0,
                        'short_PL': pnl if np.sign(units) < 0 else 0,
                        'net_position': units,
                        'PL': pnl,
                        'contracts': [item.contract],
                        'portfolio_items': [item],
                        'instrument': pos_symbol}
            
            symbol_match = instrument == pos_symbol # Product symbol match
            localSymbol_match = instrument == item.contract.localSymbol # localSymbol match
            unique_match = symbol_match or localSymbol_match
            
            if instrument is not None and unique_match:
                # The current item matches the request, append it
                key_symbol = item.contract.localSymbol if \
                    localSymbol_match else item.contract.symbol
                
                pos_dict['instrument'] = key_symbol
                
                if pos_symbol in open_positions:
                    # Portfolio item already exists, append portfolio item
                    existing_pos = open_positions[pos_symbol]
                    adjust_position(existing_pos, item, pos_dict)
                else:
                    # New position
                    open_positions[key_symbol] = Position(**pos_dict)
                
            elif instrument is None:
                # Append all positions
                if pos_symbol in open_positions:
                    # Position item already exists, append portfolio item
                    existing_pos = open_positions[pos_symbol]
                    adjust_position(existing_pos, item, pos_dict)
                else:
                    # New position
                    open_positions[pos_symbol] = Position(**pos_dict)
        
        return open_positions
    
    
    def get_summary(self) -> dict:
        """Returns account summary.
        """
        self._check_connection()
        raw_summary = self.ib.accountSummary(self.account)
        summary = self.utils.accsum_to_dict(self.account, raw_summary)
        
        return summary
    
    
    def _get_historical_data(self, instrument: str, interval: str, 
                            from_time: str, to_time: str):
        """Returns historical price data.
        """
        # self._check_connection()
        # self.ib.reqHistoricalData()
        raise NotImplementedError("This method is not available.")
        
        
    def _connect(self):
        """Connects from IB application.
        """
        self.ib.connect(host=self.host, port=self.port, clientId=self.client_id, 
                        readonly=self.read_only, account=self.account)

    
    def _disconnect(self):
        """Disconnects from IB application.
        """
        self.ib.disconnect()
        

    def _check_connection(self):
        """Checks if there is an active connection to IB. If not, will 
        attempt to reconnect.
        """
        self._refresh()
        connected = self.ib.isConnected()
        
        while not connected:
            try:
                # Try to connect
                self.ib = ib_insync.IB()
                self._connect()
            except:
                print("Connection to IB failed... trying to reconnect.")
                # Connection failed, increment client ID
                self.client_id = random.randint(0, 9999)
                
                # Sleep for a little while
                self.ib.sleep(10)
            
            # Update connection status
            connected = self.ib.isConnected()
        
    
    def _refresh(self):
        """Refreshes IB session events.
        """
        self.ib.sleep(0)
    
    
    def _get_account(self,):
        """Returns the first managed account.
        """
        self._check_connection()
        accounts = self.ib.managedAccounts()
        return accounts[0]
    
    
    def _close_position(self, order: Order, **kwargs):
        """Closes open position of symbol by placing opposing market order.
        
        Warning
        -------
        If the order instrument is for an underlying product, all contracts 
        held attributed to the underlying will be closed.
        """
        self._check_connection()
        
        symbol = order.localSymbol if order.localSymbol is not None else order.instrument
        positions = self.get_positions(instrument=symbol)
        position = positions[symbol]
        
        for item in position.portfolio_items:
            # Place opposing market order
            item_units = item.position
            action = 'BUY' if item_units < 0 else 'SELL'
            units = abs(item_units)
            IB_order = ib_insync.MarketOrder(action, units)
            contract = item.contract
            self.ib.qualifyContracts(contract)
            self.ib.placeOrder(contract, IB_order)
        
    
    def _place_market_order(self, order: Order):
        """Places a market order.
        """
        self._check_connection()
        
        # Build contract
        contract = self.utils.build_contract(order)
        
        # Create market order
        action = 'BUY' if order.direction > 0 else 'SELL'
        units = abs(order.size)
        market_order = ib_insync.MarketOrder(action, units, 
                                             orderId=self.ib.client.getReqId(),
                                             transmit=False)
        
        # Attach SL and TP orders
        orders = self._attach_auxiliary_orders(order, market_order)
        
        # Submit orders
        self._process_orders(contract, orders)
        
    
    def _place_stop_limit_order(self, order):
        """Places stop-limit order.
        """
        self._check_connection()
        
        # Build contract
        contract = self.utils.build_contract(order)
        
        # Create stop limit order
        action = 'BUY' if order.direction > 0 else 'SELL'
        units = abs(order.size)
        lmtPrice = order.order_limit_price
        stopPrice = order.order_stop_price
        IBorder = ib_insync.StopLimitOrder(action, units, lmtPrice, stopPrice, 
                                         orderId=self.ib.client.getReqId(),
                                         transmit=False)
        
        # Attach SL and TP orders
        orders = self._attach_auxiliary_orders(order, IBorder)
        
        # Submit orders
        self._process_orders(contract, orders)
    
    
    def _place_limit_order(self, order):
        """Places limit order.
        """
        self._check_connection()
        
        # Build contract
        contract = self.utils.build_contract(order)
        
        action = 'BUY' if order.direction > 0 else 'SELL'
        units = abs(order.size)
        lmtPrice = order.order_limit_price
        IBorder = ib_insync.LimitOrder(action, units, lmtPrice, 
                                     orderId=self.ib.client.getReqId(),
                                     transmit=False)
        
        # Attach SL and TP orders
        orders = self._attach_auxiliary_orders(order, IBorder)
        
        # Submit orders
        self._process_orders(contract, orders)
    
    
    def _attach_auxiliary_orders(self, order: Order, 
                                 parent_order: ib_insync.order) -> list:
        orders = [parent_order]
        
        # TP order
        if order.take_profit is not None:
            takeProfit_order = self._create_take_profit_order(order, 
                                                              parent_order.orderId)
            orders.append(takeProfit_order)
        
        # SL order
        if order.stop_loss is not None:
            stopLoss_order = self._create_stop_loss_order(order,
                                                          parent_order.orderId)
            orders.append(stopLoss_order)
        
        return orders
    
    
    def _process_orders(self, contract: ib_insync.Contract, orders: list) -> None:
        """Processes a list of orders for a given contract.
        """
        self._check_connection()
        
        # Submit orders
        for i, order in enumerate(orders):
            if i == len(orders)-1:
                # Final order; set transmit to True
                order.transmit = True
            else:
                order.transmit = False
            self.ib.placeOrder(contract, order)
    
    
    def _convert_to_oca(self, orders: list, oca_group: str = None, 
                        oca_type: int = 1) -> list:
        """Converts a list of Orders to One Cancels All group of orders.

        Parameters
        ----------
        orders : list
            A list of orders.

        Returns
        -------
        oca_orders : list
            The orders modified to be in a OCA group.
        """
        self._check_connection()
        
        if oca_group is None:
            oca_group = f'OCA_{self.ib.client.getReqId()}'
        
        oca_orders = self.ib.oneCancelsAll(orders, oca_group, oca_type)
        return oca_orders
    
    
    def _create_take_profit_order(self, order: Order, parentId: int):
        """Constructs a take profit order.
        """
        quantity = order.size
        takeProfitPrice = order.take_profit
        action = 'BUY' if order.direction < 0 else 'SELL'
        takeProfit_order = ib_insync.LimitOrder(action, 
                                                quantity, 
                                                takeProfitPrice,
                                                orderId=self.ib.client.getReqId(),
                                                transmit=False,
                                                parentId=parentId)
        return takeProfit_order
    
    
    def _create_stop_loss_order(self, order: Order, parentId: int):
        """Constructs a stop loss order.
        """
        # TODO - add support for trailing SL
        quantity = order.size
        stopLossPrice = order.stop_loss
        action = 'BUY' if order.direction < 0 else 'SELL'
        stopLoss_order = ib_insync.StopOrder(action, 
                                             quantity, 
                                             stopLossPrice,
                                             orderId=self.ib.client.getReqId(),
                                             transmit=True,
                                             parentId=parentId)
        return stopLoss_order
    



import os
import numpy as np
import pandas as pd
from datetime import datetime


class BrokerUtils:
    def __init__(self):
        pass
    
    
    def __repr__(self):
        return 'AutoTrader Broker Utilities'
    
    
    def __str__(self):
        return 'AutoTrader Broker Utilities'
    
    
    def response_to_df(self, response: pd.DataFrame):
        """Function to convert api response into a pandas dataframe.
        """
        candles = response.body["candles"]
        times = []
        close_price, high_price, low_price, open_price = [], [], [], []
        
        for candle in candles:
            times.append(candle.time)
            close_price.append(float(candle.mid.c))
            high_price.append(float(candle.mid.h))
            low_price.append(float(candle.mid.l))
            open_price.append(float(candle.mid.o))
        
        dataframe = pd.DataFrame({"Open": open_price, "High": high_price, "Low": low_price, "Close": close_price})
        dataframe.index = pd.to_datetime(times)
        
        return dataframe
    
    
    def truncate(self, f: float, n: int):
        """Truncates a float f to n decimal places without rounding. 
        """
        s = '{}'.format(f)
        
        if 'e' in s or 'E' in s:
            return '{0:.{1}f}'.format(f, n)
        i, p, d = s.partition('.')
        
        return '.'.join([i, (d+'0'*n)[:n]])
    
    
    def get_pip_ratio(self, pair):
        """Function to return pip value ($/pip) of a given forex pair.
        """
        # TODO - implications on non FX?
        if 'JPY' in pair:
            pip_value = 1e-2
        else:
            pip_value = 1e-4
        
        return pip_value
    
    
    def get_size(self, pair: str, amount_risked: float, price: float, 
                 stop_price: float, HCF: float, 
                 stop_distance: float = None) -> float:
        """Calculate position size based on account balance and risk profile.
        
        References
        ----------
        https://www.babypips.com/tools/position-size-calculator
        """
        if stop_price is None and stop_distance is None:
            # No stop loss being used, instead risk portion of account
            units               = amount_risked/(HCF*price)
        else:
            pip_value           = self.get_pip_ratio(pair)
            
            if stop_price is None:
                pip_stop_distance = stop_distance
            else:
                pip_stop_distance = abs(price - stop_price) / pip_value
            
            
            if pip_stop_distance == 0:
                units           = 0
            else:
                quote_risk      = amount_risked / HCF
                price_per_pip   = quote_risk / pip_stop_distance
                units           = price_per_pip / pip_value
        
        return units
    
    
    def check_precision(self, pair, original_stop, original_take):
        """Modify stop/take based on pair for required ordering precision. 
        """
        if pair[-3:] == 'JPY':
            N = 3
        else:
            N = 5
        
        take_price      = float(self.truncate(original_take, N))
        stop_price      = float(self.truncate(original_stop, N))
        
        return stop_price, take_price
    
    
    def interval_to_seconds(self, interval):
        """Converts the interval to time in seconds.
        """
        letter = interval[0]
        
        if len(interval) > 1:
            number = float(interval[1:])
        else:
            number = 1
        
        conversions = {'S': 1,
                       'M': 60,
                       'H': 60*60,
                       'D': 60*60*24
                       }
        
        my_int = conversions[letter] * number
        
        return my_int
    
    
    def write_to_order_summary(self, order, filepath: str):
        """Writes order details to summary file.
        """
        # Check if file exists already, if not, create
        if not os.path.exists(filepath):
            f = open(filepath, "w")
            f.write("order time, strategy, granularity, order_type, instrument, order_size, ")
            f.write("trigger_price, stop_loss, take_profit\n")
            f.close()
        
        order_time = order.order_time 
        strategy = order.strategy
        order_type = order.order_type
        instrument = order.instrument
        size = order.size
        trigger_price = order.order_price
        stop_loss = order.stop_loss
        take_profit = order.take_profit
        granularity = order.granularity
        
        f = open(filepath, "a")
        f.write("{}, {}, {}, {}, {}, {}, {}, {}, {}\n".format(order_time, strategy, 
              granularity, order_type, instrument, size, trigger_price, stop_loss, 
              take_profit))
        f.close()
        
    
    def check_dataframes(self, df_1: pd.DataFrame, df_2: pd.DataFrame):
        """Checks dataframe lengths and corrects if necessary.
        """
        if len(df_1) < len(df_2):
            new_df_1 = self.fix_dataframe(df_2, df_1)
            new_df_2 = df_2
        elif len(df_1) > len(df_2):
            new_df_2 = self.fix_dataframe(df_1, df_2)
            new_df_1 = df_1
            
        else:
            new_df_1 = df_1
            new_df_2 = df_2
        
        return new_df_1, new_df_2
    
    
    def fix_dataframe(self, df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
        """Ensures that the quote data and data dataframes are the same
        lenght.
        """
        # Would be good to check which one is shorter, which is longer, then 
        # return both with corrections
        
        i1 = list(df1.index)
        i2 = list(df2.index)
        new_indices = list(set(i1) - set(i2))
        
        new_df = df2
        
        for index in new_indices:
            
            df_row = df1.copy()[df1.index == index]
            
            df_row.Open = None
            df_row.High = None
            df_row.Low = None
            df_row.Close = None
            
            new_df = new_df.append(df_row)
        
        new_df = new_df.sort_index()
        new_df = new_df.interpolate()
        
        return new_df
    
    
    def trade_summary(self, trades: dict = None, orders: dict = None, 
                      instrument: str = None) -> pd.DataFrame:
        """Creates backtest trade summary dataframe.
        """
        if trades is not None:
            iter_dict = trades
        else:
            iter_dict = orders
        
        iter_dict = {} if iter_dict is None else iter_dict 
        
        product = []
        status = []
        ids = []
        times_list = []
        order_price = []
        size = []
        direction = []
        stop_price = []
        take_price = []
        
        if trades is not None:
            entry_time = []
            fill_price = []
            profit = []
            portfolio_balance = []
            exit_times = []
            exit_prices = []
            trade_duration = []
            fees = []
        
        for ID, item in iter_dict.items():
            product.append(item.instrument)
            status.append(item.status)
            ids.append(item.id)
            size.append(item.size)
            direction.append(item.direction)
            times_list.append(item.order_time)
            order_price.append(item.order_price)
            stop_price.append(item.stop_loss)
            take_price.append(item.take_profit)
        
        if trades is not None:
            for trade_id, trade in iter_dict.items():
                entry_time.append(trade.time_filled)
                fill_price.append(trade.fill_price)
                profit.append(trade.profit)
                portfolio_balance.append(trade.balance)
                exit_times.append(trade.exit_time)
                exit_prices.append(trade.exit_price)
                fees.append(trade.fees)
                if trade.status == 'closed':
                    if type(trade.exit_time) == str:
                        exit_dt = datetime.strptime(trade.exit_time, "%Y-%m-%d %H:%M:%S%z")
                        entry_dt = datetime.strptime(trade.time_filled, "%Y-%m-%d %H:%M:%S%z")
                        trade_duration.append(exit_dt.timestamp() - entry_dt.timestamp())
                    elif isinstance(trade.exit_time, pd.Timestamp):
                        trade_duration.append((trade.exit_time - trade.time_filled).total_seconds())
                    else:
                        trade_duration.append(trade.exit_time.timestamp() - 
                                              trade.time_filled.timestamp())
                else:
                    trade_duration.append(None)
                
            dataframe = pd.DataFrame({"instrument": product,
                                      "status": status,
                                      "ID": ids, 
                                      "order_price": order_price,
                                      "order_time": times_list,
                                      "fill_time": entry_time,
                                      "fill_price": fill_price, "size": size,
                                      "direction": direction,
                                      "stop_loss": stop_price, "take_profit": take_price,
                                      "profit": profit, "balance": portfolio_balance,
                                      "exit_time": exit_times, "exit_price": exit_prices,
                                      "trade_duration": trade_duration,
                                      "fees": fees},
                                     index = pd.to_datetime(entry_time))
            
            # Fill missing values for balance
            dataframe.balance.fillna(method='ffill', inplace=True)
            
        else:
            dataframe = pd.DataFrame({"instrument": product,
                                      "status": status,
                                      "ID": ids, 
                                      "order_price": order_price,
                                      "order_time": times_list,
                                      "size": size,
                                      "direction": direction,
                                      "stop_loss": stop_price, 
                                      "take_profit": take_price},
                                     index = pd.to_datetime(times_list))
            
        dataframe = dataframe.sort_index()
        
        # Filter by instrument
        if instrument is not None:
            dataframe = dataframe[dataframe['instrument'] == instrument]
        
        return dataframe
    
    
    def get_streaks(self, trade_summary):
        """Calculates longest winning and losing streaks from trade summary. 
        """
        profit_list = trade_summary[trade_summary['status']=='closed'].profit.values
        longest_winning_streak = 1
        longest_losing_streak = 1
        streak = 1
        
        for i in range(1, len(profit_list)):
            if np.sign(profit_list[i]) == np.sign(profit_list[i-1]):
                streak += 1
                
                if np.sign(profit_list[i]) > 0:
                    # update winning streak
                    longest_winning_streak  = max(longest_winning_streak, streak)
                else:
                    # Update losing 
                    longest_losing_streak   = max(longest_losing_streak, streak)
    
            else:
                streak = 1
        
        return longest_winning_streak, longest_losing_streak




import os
from macd_strategy import SimpleMACD
from autotrader.autotrader import AutoTrader

def test_macd_backtest():
    config = {'NAME': 'MACD Strategy',
              'MODULE': 'macd_strategy',
              'CLASS': 'SimpleMACD',
              'INTERVAL': 'H4',
              'PERIOD': 300,
              'RISK_PC': 1.5,
              'SIZING': 'risk',
              'PARAMETERS': {'ema_period': 200,
                             'MACD_fast': 5,
                             'MACD_slow': 19,
                             'MACD_smoothing': 9,
                             'RR': 1.5},
              'WATCHLIST': ['EUR_USD'],}
    home_dir = os.path.abspath(os.path.dirname(__file__))
    
    at = AutoTrader()
    at.configure(verbosity=2, show_plot=True)
    at.add_strategy(config_dict=config, strategy=SimpleMACD)
    at.plot_settings(show_cancelled=False)
    at.add_data({'EUR_USD': 'EUR_USD_H4.csv'}, 
                data_directory=os.path.join(home_dir, 'data'))
    at.backtest(start = '1/1/2015', end = '1/3/2022',
                initial_balance=1000, leverage=30,
                spread=0.5, commission=0.005, hedging=True)
    at.run()
    bot = at.get_bots_deployed()
    bt_results = at.analyse_backtest(bot)
    
    # Test backtest results
    assert bt_results['no_trades'] == 229, "Incorrect number of trades"
    assert round(bt_results['all_trades']['ending_balance'], 3) == 774.5, "Incorrect ending balance"
    assert bt_results['long_trades']['no_trades'] == 95, "Incorrect number of long trades"
    assert bt_results['short_trades']['no_trades'] == 134, "Incorrect number of short trades"


def test_multibot_macd_backtest():
    config = {'NAME': 'MACD Strategy',
              'MODULE': 'macd_strategy',
              'CLASS': 'SimpleMACD',
              'INTERVAL': 'H4',
              'PERIOD': 300,
              'RISK_PC': 1.5,
              'SIZING': 'risk',
              'PARAMETERS': {'ema_period': 200,
                              'MACD_fast': 5,
                              'MACD_slow': 19,
                              'MACD_smoothing': 9,
                              'RR': 1.5},
              'WATCHLIST': ['EUR_USD', 'EUR_USD2'],}
    home_dir = os.path.abspath(os.path.dirname(__file__))
    
    at = AutoTrader()
    at.configure(verbosity=2, show_plot=True)
    at.add_strategy(config_dict=config, strategy=SimpleMACD)
    at.plot_settings(show_cancelled=False)
    at.add_data({'EUR_USD': 'EUR_USD_H4.csv',
                  'EUR_USD2': 'EUR_USD_H4.csv'}, 
                data_directory=os.path.join(home_dir, 'data'))
    at.backtest(start = '1/1/2015', end = '1/3/2022',
                initial_balance=1000, leverage=30,
                spread=0.5, commission=0.005, hedging=True)
    at.run()
    bots = at.get_bots_deployed()
    EU1bot = bots['EUR_USD']
    bt_results = at.multibot_backtest_results
    
    assert round(bt_results['win_rate'][0],5) == 38.59649
    assert round(bt_results['win_rate'][1],5) == 39.28571
    assert bt_results['no_long'][0] == 95, "Incorrect number of long trades"
    assert bt_results['no_long'][1] == 93, "Incorrect number of long trades"
    assert bt_results['no_short'][0] == 133, "Incorrect number of short trades"
    assert bt_results['no_short'][1] == 131, "Incorrect number of short trades"
    assert round(EU1bot.backtest_summary['account_history']['balance'][-1], 5) == 586.76734, "Incorrect ending balance"




from autotrader.brokers.broker_utils import BrokerUtils

class Utils(BrokerUtils):
    def __init__(self):
        pass
    
    
    def __repr__(self):
        return 'AutoTrader Broker Utilities'
    
    
    def __str__(self):
        return 'AutoTrader Broker Utilities'
    
    


import v20
import ib_insync
import pandas as pd
import yfinance as yf
from typing import Union
from autotrader.brokers.trading import Order
from datetime import datetime, timedelta, timezone
from autotrader.brokers.ib.utils import Utils as IB_Utils


class GetData:
    """GetData class to retrieve price data.

    Attributes
    ----------
    home_curreny : str
        the home currency of the account (used for retrieving quote data)
    
    allow_dancing_bears : bool
        Allow incomplete candlesticks in data retrieval.

    Methods
    -------
    oanda()
        Retrieves historical price data of a instrument from Oanda.
    yahoo()
        Retrieves historical price data from Yahoo Finance via yfinance.
    local()
        Reads local price data.
    """
    
    def __init__(self, broker_config: dict = None, 
                 allow_dancing_bears: bool = False,
                 home_currency: str = None) -> None:
        """Instantiates GetData.

        Parameters
        ----------
        broker_config : dict, optional
            The configuration dictionary for the broker to be used. The 
            default is None.
        allow_dancing_bears : bool, optional
            A flag to allow incomplete bars to be returned in the data. The 
            default is False.
        home_currency : str, optional
            The home currency to use when fetching quote data. The default 
            is None.

        Returns
        -------
        None
            GetData will be instantiated and ready to fetch price data.

        """
        if broker_config is not None:
            if broker_config['data_source'] == 'OANDA':
                API = broker_config["API"]
                ACCESS_TOKEN = broker_config["ACCESS_TOKEN"]
                port = broker_config["PORT"]
                self.api = v20.Context(hostname=API, token=ACCESS_TOKEN, port=port)
            
            elif broker_config['data_source'] == 'IB':
                host = broker_config['host']
                port = broker_config['port'] 
                client_id = broker_config['clientID'] + 1
                read_only = broker_config['read_only']
                account = broker_config['account']
                
                self.ibapi = ib_insync.IB()
                self.ibapi.connect(host=host, port=port, clientId=client_id, 
                                   readonly=read_only, account=account)
            
        self.allow_dancing_bears = allow_dancing_bears
        self.home_currency = home_currency
        

    def oanda(self, instrument: str, granularity: str, count: int = None, 
              start_time: datetime = None, end_time: datetime = None) -> pd.DataFrame:
        """Retrieves historical price data of a instrument from Oanda v20 API.

        Parameters
        ----------
        instrument : str
            The instrument to fetch data for.
        granularity : str
            The candlestick granularity (eg. "M15", "H4", "D").
        count : int, optional
            The number of candles to fetch (maximum 5000). The default is None.
        start_time : datetime, optional
            The data start time. The default is None.
        end_time : datetime, optional
            The data end time. The default is None.

        Returns
        -------
        data : DataFrame
            The price data, as an OHLC DataFrame.
            
        Notes
        -----
            If a candlestick count is provided, only one of start time or end
            time should be provided. If neither is provided, the N most
            recent candles will be provided. If both are provided, the count
            will be ignored, and instead the dates will be used.
        
        Examples
        --------
        >>> data = GetData.oanda("EUR_USD", granularity="M15", 
                                 start_time=from_dt, end_time=to_dt)
            
        >>> data = GetData.oanda("EUR_USD", granularity="M15",
                                 start_time=from_dt, count=2110)
        
        >>> data = GetData.oanda("EUR_USD", granularity="M15",
                                 end_time=to_dt, count=2110)
        
        >>> data = GetData.oanda("EUR_USD", granularity="M15", 
                                 count=2110)
        """
        if count is not None:
            # either of count, start_time+count, end_time+count (or start_time+end_time+count)
            # if count is provided, count must be less than 5000
            if start_time is None and end_time is None:
                # fetch count=N most recent candles
                response = self.api.instrument.candles(instrument,
                                                       granularity = granularity,
                                                       count = count)
                data = self.response_to_df(response)
                
            elif start_time is not None and end_time is None:
                # start_time + count
                from_time = start_time.timestamp()
                response = self.api.instrument.candles(instrument,
                                                       granularity = granularity,
                                                       count = count,
                                                       fromTime = from_time)
                data = self.response_to_df(response)
            
            elif end_time is not None and start_time is None:
                # end_time + count
                to_time = end_time.timestamp()
                response = self.api.instrument.candles(instrument,
                                             granularity = granularity,
                                             count = count,
                                             toTime = to_time)
                data = self.response_to_df(response)
                
            else:
                # start_time+end_time+count
                # print("Warning: ignoring count input since start_time and",
                #        "end_time times have been specified.")
                from_time = start_time.timestamp()
                to_time = end_time.timestamp()
            
                # try to get data 
                response = self.api.instrument.candles(instrument,
                                                       granularity = granularity,
                                                       fromTime = from_time,
                                                       toTime = to_time)
                
                # If the request is rejected, max candles likely exceeded
                if response.status != 200:
                    data = self._get_extended_oanda_data(instrument,
                                                         granularity,
                                                         from_time,
                                                         to_time)
                else:
                    data = self.response_to_df(response)
                
        else:
            # count is None
            # Assume that both start_time and end_time have been specified.
            from_time = start_time.timestamp()
            to_time = end_time.timestamp()
            
            # try to get data 
            response = self.api.instrument.candles(instrument,
                                                   granularity = granularity,
                                                   fromTime = from_time,
                                                   toTime = to_time)
            
            # If the request is rejected, max candles likely exceeded
            if response.status != 200:
                data = self._get_extended_oanda_data(instrument,
                                                     granularity,
                                                     from_time,
                                                     to_time)
            else:
                data = self.response_to_df(response)

        return data
    
    
    def oanda_liveprice(self, order: Order, **kwargs) -> dict:
        """Returns current price (bid+ask) and home conversion factors.
        """
        response = self.api.pricing.get(accountID = self.ACCOUNT_ID, 
                                        instruments = order.instrument)
        ask = response.body["prices"][0].closeoutAsk
        bid = response.body["prices"][0].closeoutBid
        negativeHCF = response.body["prices"][0].quoteHomeConversionFactors.negativeUnits
        positiveHCF = response.body["prices"][0].quoteHomeConversionFactors.positiveUnits
    
        price = {"ask": ask,
                 "bid": bid,
                 "negativeHCF": negativeHCF,
                 "positiveHCF": positiveHCF}
    
        return price
    
    
    def _get_extended_oanda_data(self, instrument, granularity, from_time, to_time):
        """Returns historical data between a date range."""
        max_candles = 5000
        
        my_int = self._granularity_to_seconds(granularity, 'oanda')
        end_time = to_time - my_int
        partial_from = from_time
        response = self.api.instrument.candles(instrument,
                                               granularity = granularity,
                                               fromTime = partial_from,
                                               count = max_candles)
        data = self.response_to_df(response)
        last_time = data.index[-1].timestamp()
        
        while last_time < end_time:
            partial_from = last_time
            response = self.api.instrument.candles(instrument,
                                                   granularity = granularity,
                                                   fromTime = partial_from,
                                                   count = max_candles)
            
            partial_data = self.response_to_df(response)
            data = data.append(partial_data)
            last_time = data.index[-1].timestamp()
            
        return data
    
    
    def _oanda_quote_data(self, data: pd.DataFrame, pair: str, granularity: str, 
                         start_time: datetime, end_time: datetime):
        """Function to retrieve price conversion data.
        """
        base_currency = pair[:3]
        quote_currency = pair[-3:]
        
        if self.home_currency is None or quote_currency == self.home_currency:
            # Use data as quote data
            quote_data = data
            
        else:
            if self.home_currency == base_currency:
                # Invert data to get quote data
                quote_data = 1/data
                
            else:
                # Try download quote data
                try:
                    conversion_pair = self.home_currency + "_" + quote_currency
                    if conversion_pair != pair:
                        # Do not re-download the same data
                        quote_data = self.oanda(instrument  = conversion_pair,
                                                granularity = granularity,
                                                start_time  = start_time,
                                                end_time    = end_time)
                    else:
                        quote_data = data
                        
                except:
                    # Download failed, revert to original data
                    quote_data = data
        
        return quote_data
    
    
    def _check_oanda_response(self, response):
        """Placeholder method to check Oanda API response.
        """
        if response.status != 200:
            print(response.reason)
    
    
    def response_to_df(self, response):
        """Function to convert api response into a pandas dataframe.
        """
        candles = response.body["candles"]
        times = []
        close_price, high_price, low_price, open_price = [], [], [], []
        
        for candle in candles:
            if self.allow_dancing_bears:
                times.append(candle.time)
                close_price.append(float(candle.mid.c))
                high_price.append(float(candle.mid.h))
                low_price.append(float(candle.mid.l))
                open_price.append(float(candle.mid.o))
                
            else:
                if candle.complete:
                    times.append(candle.time)
                    close_price.append(float(candle.mid.c))
                    high_price.append(float(candle.mid.h))
                    low_price.append(float(candle.mid.l))
                    open_price.append(float(candle.mid.o))
        
        dataframe = pd.DataFrame({"Open": open_price, "High": high_price, "Low": low_price, "Close": close_price})
        dataframe.index = pd.to_datetime(times)
        dataframe.drop_duplicates(inplace=True)
        
        return dataframe
    
    
    @staticmethod
    def _granularity_to_seconds(granularity: str, feed: str):
        """Converts the granularity to time in seconds.
        """
        if feed.lower() == 'oanda':
            allowed_granularities = ('S5', 'S10', 'S15', 'S30',
                                       'M1', 'M2', 'M4', 'M5', 'M10', 'M15', 'M30',
                                       'H1', 'H2', 'H3', 'H4', 'H6', 'H8', 'H12',
                                       'D', 'W', 'M')
            
            if granularity not in allowed_granularities:
                raise Exception(f"Invalid granularity '{granularity}' for "+\
                                "{feed} data feed.")
            
            letter = granularity[0]
            
            if len(granularity) > 1:
                number = float(granularity[1:])
            else:
                number = 1
            
            conversions = {'S': 1,
                           'M': 60,
                           'H': 60*60,
                           'D': 60*60*24}
            
            my_int = conversions[letter] * number
            
        elif feed.lower() == 'yahoo':
            # Note: will not work for week or month granularities
            
            letter = granularity[-1]
            number = float(granularity[:-1])
            
            conversions = {'m': 60,
                           'h': 60*60,
                           'd': 60*60*24}
            
            my_int = conversions[letter] * number
        
        return my_int
    

    def yahoo(self, instrument: str, granularity: str = None, count: int = None, 
              start_time: str = None, end_time: str = None) -> pd.DataFrame:
        """Retrieves historical price data from yahoo finance. 

        Parameters
        ----------
        instrument : str
            Ticker to dowload data for.
        granularity : str, optional
            The candlestick granularity. The default is None.
        count : int, optional
            The number of bars to fetch. The default is None.
        start_time : str, optional
            The start time as YYYY-MM-DD string or datetime object. The default 
            is None.
        end_time : str, optional
            The end_time as YYYY-MM-DD string or datetime object. The default 
            is None.

        Returns
        -------
        data : pd.DataFrame
            The price data, as an OHLC DataFrame.

        Notes
        -----
        If you are encountering a JSON error when using the yahoo finance API,
        try updating by running: pip install yfinance --upgrade --no-cache-dir

        Intraday data cannot exceed 60 days.
        """
        
        if count is not None and start_time is None and end_time is None:
            # Convert count to start and end dates (assumes end=now)
            end_time = datetime.now()
            start_time = end_time - timedelta(seconds=self._granularity_to_seconds(granularity, 'yahoo')*count)
        
        data = yf.download(tickers  = instrument, 
                           start    = start_time, 
                           end      = end_time,
                           interval = granularity)
        
        # Add UTC timezone
        data.index = data.index.tz_localize(timezone.utc)
        
        return data
    
    
    def yahoo_liveprice(self,):
        raise Exception("Live price is not available from yahoo API.")
        
    
    def _yahoo_quote_data(self, data: pd.DataFrame, pair: str, interval: str,
                         from_date: datetime, to_date: datetime):
        """Returns nominal price data - quote conversion not supported for 
        Yahoo finance API.
        """
        return data
    
    
    def _check_IB_connection(self):
        """Checks if there is an active connection to IB.
        """
        self.ibapi.sleep(0)
        connected = self.ibapi.isConnected()
        if not connected:
            raise ConnectionError("No active connection to IB.")
    
    
    def ib(self, instrument: str, granularity: str, count: int,
           start_time: datetime = None, end_time: datetime = None,
           order: Order = None, durationStr: str = '10 mins', **kwargs) -> pd.DataFrame:
        """

        Parameters
        ----------
        instrument : str
            The product being traded.
        granularity : str
            The granularity of the price bars.
        count : int
            The number of bars to fetch.
        start_time : datetime, optional
            The data start time. The default is None.
        end_time : datetime, optional
            The data end time. The default is None.
        order : Order, optional
            The order object. The default is None.

        Raises
        ------
        NotImplementedError
            DESCRIPTION.

        Returns
        -------
        df : TYPE
            DESCRIPTION.
        
        Warnings
        --------
        This method is not recommended due to its high API poll rate.
        
        References
        ----------
        https://ib-insync.readthedocs.io/api.html?highlight=reqhistoricaldata#
        """
        raise NotImplementedError("Historical market data from IB is not yet supported.")
        # TODO - implement
        contract = IB_Utils.build_contract(order)
        
        dt = ''
        barsList = []
        while True:
            bars = self.ibapi.reqHistoricalData(
                contract,
                endDateTime=dt,
                durationStr=durationStr,
                barSizeSetting=granularity,
                whatToShow='MIDPOINT',
                useRTH=True,
                formatDate=1)
            if not bars:
                break
            barsList.append(bars)
            dt = bars[0].date
        
        # Convert bars to DataFrame
        allBars = [b for bars in reversed(barsList) for b in bars]
        df = self.ibapi.util.df(allBars)
        return df
    
    
    def ib_liveprice(self, order: Order, snapshot: bool = False, **kwargs) -> dict:
        """Returns current price (bid+ask) and home conversion factors.
        
        Parameters
        ----------
        order: Order
            The AutoTrader Order.
        snapshot : bool, optional
            Request a snapshot of the price. The default is False.

        Returns
        -------
        dict
            A dictionary containing the bid and ask prices.
        
        """
        self._check_IB_connection()
        contract = IB_Utils.build_contract(order)
        self.ibapi.qualifyContracts(contract)
        ticker = self.ibapi.reqMktData(contract, snapshot=snapshot)
        while ticker.last != ticker.last: self.ibapi.sleep(1)
        self.ibapi.cancelMktData(contract)
        price = {"ask": ticker.ask,
                 "bid": ticker.bid,
                 "negativeHCF": 1,
                 "positiveHCF": 1,}
        return price
    
    
    @staticmethod
    def _pseduo_liveprice(last: float, quote_price: float = None) -> dict:
        """Returns an artificial live bid and ask price, plus conversion 
        factors.

        Parameters
        ----------
        last : float
            The last price of the product being traded.
        quote_price : float, optional
            The quote price of the product being traded against the account
            home currency. The default is None.

        Returns
        -------
        dict
            DESCRIPTION.

        """
        # TODO - build bid/ask spread into here, review virtual broker
        if quote_price is not None:
            # Use quote price to determine HCF
            if last == quote_price:
                # Quote currency matches account home currency
                negativeHCF = 1
                positiveHCF = 1
            else:
                # Quote data
                negativeHCF = quote_price
                positiveHCF = quote_price
                
        else:
            # No quote price provided
            negativeHCF = 1
            positiveHCF = 1
        
        price = {"ask": last,
                 "bid": last,
                 "negativeHCF": negativeHCF,
                 "positiveHCF": positiveHCF}
        
        return price
    
    
    @staticmethod
    def local(filepath: str, start_date: Union[str, datetime] = None, 
              end_date: Union[str, datetime] = None, utc: bool = True) -> pd.DataFrame:
        """Read local price data.

        Parameters
        ----------
        filepath : str
            The absolute filepath of the local price data.
        start_date : str | datetime, optional
            The data start date. The default is None.
        end_date : str | datetime, optional
            The data end data. The default is None.
        utc : bool, optional
            Localise data to UTC. The default is True.

        Returns
        -------
        data : pd.DataFrame
            The price data, as an OHLC DataFrame.
        """
        data = pd.read_csv(filepath, index_col = 0)
        data.index = pd.to_datetime(data.index, utc=utc)
        
        if start_date is not None and end_date is not None:
            # Filter by date range
            data = GetData._check_data_period(data, start_date, 
                                              end_date)
            
        return data
    
    
    @staticmethod
    def _check_data_period(data: pd.DataFrame, start_date: datetime, 
                           end_date: datetime) -> pd.DataFrame:
        """Checks and returns the dataset matching the backtest start and 
        end dates (as close as possible).
        """
        return data[(data.index >= start_date) & (data.index <= end_date)]
